"""Principal Component Analysis on Manifolds.

Lead author: Nina Miolane.
"""

import numbers
from math import log

from scipy.special import gammaln
from sklearn.decomposition._base import _BasePCA
from sklearn.utils.extmath import stable_cumsum, svd_flip

import geomstats.backend as gs
from geomstats.geometry._hyperbolic import _Hyperbolic
from geomstats.geometry.hyperbolic import Hyperbolic
from geomstats.geometry.matrices import Matrices
from geomstats.geometry.symmetric_matrices import SymmetricMatrices
from geomstats.learning.exponential_barycenter import ExponentialBarycenter
from geomstats.learning.frechet_mean import FrechetMean


def _assess_dimension_(spectrum, rank, n_samples, n_features):
    """Compute the likelihood of a rank ``rank`` dataset.

    The dataset is assumed to be embedded in gaussian noise of shape(n,
    dimf) having spectrum ``spectrum``.

    Parameters
    ----------
    spectrum : array of shape (n)
        Data spectrum.
    rank : int
        Tested rank value.
    n_samples : int
        Number of samples.
    n_features : int
        Number of features.

    Returns
    -------
    ll : float
        Log-likelihood.

    Notes
    -----
    This implements the method of `Thomas P. Minka:
    Automatic Choice of Dimensionality for PCA. NIPS 2000: 598-604`
    """
    if rank > len(spectrum):
        raise ValueError("The tested rank cannot exceed the rank of the dataset")

    pu = -rank * log(2.0)
    for i in range(rank):
        pu += gammaln((n_features - i) / 2.0) - log(gs.pi) * (n_features - i) / 2.0

    pl = gs.sum(gs.log(spectrum[:rank]))
    pl = -pl * n_samples / 2.0

    if rank == n_features:
        pv = 0
        v = 1
    else:
        v = gs.sum(spectrum[rank:]) / (n_features - rank)
        pv = -gs.log(v) * n_samples * (n_features - rank) / 2.0

    m = n_features * rank - rank * (rank + 1.0) / 2.0
    pp = log(2.0 * gs.pi) * (m + rank + 1.0) / 2.0

    pa = 0.0
    spectrum_ = spectrum.copy()
    spectrum_[rank:n_features] = v
    for i in range(rank):
        for j in range(i + 1, len(spectrum)):
            pa += log(
                (spectrum[i] - spectrum[j]) * (1.0 / spectrum_[j] - 1.0 / spectrum_[i])
            ) + log(n_samples)

    ll = pu + pl + pv + pp - pa / 2.0 - rank * log(n_samples) / 2.0

    return ll


def _infer_dimension_(spectrum, n_samples, n_features):
    """Infer the dimension of a dataset of shape (n_samples, n_features).

    The dataset is described by its spectrum `spectrum`.
    """
    n_spectrum = len(spectrum)
    ll = gs.empty(n_spectrum)
    for rank in range(n_spectrum):
        ll[rank] = _assess_dimension_(spectrum, rank, n_samples, n_features)
    return ll.argmax()


class TangentPCA(_BasePCA):
    r"""Tangent Principal component analysis (tPCA).

    Linear dimensionality reduction using
    Singular Value Decomposition of the
    Riemannian Log of the data at the tangent space
    of the Frechet mean.

    Parameters
    ----------
    space : Manifold
        Equipped manifold.
    n_components : int
        Number of principal components.
        Optional, default: None.

    Notes
    -----
    * Required geometry methods: `exp`, `log`.
    * If `base_point=None`, also requires `FrechetMean` required methods.
    * Lie groups can be used without a metric, but `base_point` or `mean_estimator`
      need to be specified.
    """

    def __init__(
        self,
        space,
        n_components=None,
        copy=True,
        whiten=False,
        tol=0.0,
        iterated_power="auto",
        random_state=None,
    ):
        self.space = space
        self.n_components = n_components
        self.copy = copy
        self.whiten = whiten
        self.tol = tol
        self.iterated_power = iterated_power
        self.random_state = random_state

        if hasattr(self.space, "metric"):
            self.mean_estimator = FrechetMean(space)
        else:
            self.mean_estimator = ExponentialBarycenter(space)

        self.base_point_ = None

    @property
    def _geometry(self):
        """Object where `exp` and `log` are defined."""
        if hasattr(self.space, "metric"):
            return self.space.metric

        return self.space

    def fit(self, X, y=None, base_point=None):
        """Fit the model with X.

        Parameters
        ----------
        X : array-like, shape=[..., n_features]
            Training data, where n_samples is the number of samples
            and n_features is the number of features.
        y : Ignored (Compliance with scikit-learn interface)
        base_point : array-like, shape=[..., n_features], optional
            Point at which to perform the tangent PCA
            Optional, default to Frechet mean if None.

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        self._fit(X, base_point=base_point)
        return self

    def fit_transform(self, X, y=None, base_point=None):
        """Fit the model with X and apply the dimensionality reduction on X.

        Parameters
        ----------
        X : array-like, shape=[..., n_features]
            Training data, where n_samples is the number of samples
            and n_features is the number of features.
        y : Ignored (Compliance with scikit-learn interface)
        base_point : array-like, shape=[..., n_features]
            Point at which to perform the tangent PCA
            Optional, default to Frechet mean if None.

        Returns
        -------
        X_new : array-like, shape=[..., n_components]
            Projected data.
        """
        U, S, _ = self._fit(X, base_point=base_point)

        U = U[:, : self.n_components_]

        U *= S[: self.n_components_]
        return U

    def transform(self, X, y=None):
        """Project X on the principal components.

        Parameters
        ----------
        X : array-like, shape=[..., n_features]
            Data, where n_samples is the number of samples
            and n_features is the number of features.
        y : Ignored (Compliance with scikit-learn interface)

        Returns
        -------
        X_new : array-like, shape=[..., n_components]
            Projected data.
        """
        tangent_vecs = self._geometry.log(X, base_point=self.base_point_)
        if self.space.point_ndim == 2:
            if (
                gs.all(Matrices.is_square(tangent_vecs))
                and Matrices.is_symmetric(tangent_vecs).all()
            ):
                X = SymmetricMatrices.basis_representation(tangent_vecs)
            else:
                X = gs.reshape(tangent_vecs, (len(X), -1))
        else:
            X = tangent_vecs
        X = X - self.mean_
        X_transformed = gs.matmul(X, gs.transpose(self.components_))
        return X_transformed

    def inverse_transform(self, X):
        """Low-dimensional reconstruction of X.

        The reconstruction will match X_original whose transform would be X
        if `n_components=min(n_samples, n_features)`.

        Parameters
        ----------
        X : array-like, shape=[..., n_components]
            New data, where n_samples is the number of samples
            and n_components is the number of components.

        Returns
        -------
        X_original : array-like, shape=[..., n_features]
            Original data.
        """
        scores = self.mean_ + gs.matmul(X, self.components_)

        if self.space.point_ndim > 1:
            if gs.all(Matrices.is_square(self.base_point_)) and gs.all(
                Matrices.is_symmetric(self.base_point_)
            ):
                scores = SymmetricMatrices(
                    self.base_point_.shape[-1]
                ).matrix_representation(scores)
            else:
                dim = self.base_point_.shape[-1]
                scores = gs.reshape(scores, (len(scores), dim, dim))
        return self._geometry.exp(scores, self.base_point_)

    def _fit(self, X, base_point=None):
        """Fit the model by computing full SVD on X.

        Parameters
        ----------
        X : array-like, shape=[..., n_features]
            Training data, where n_samples is the number of samples
            and n_features is the number of features.
        y : Ignored (Compliance with scikit-learn interface)
        base_point : array-like, shape=[..., n_features]
            Point at which to perform the tangent PCA.
            Optional, default to Frechet mean if None.

        Returns
        -------
        U, S, V : array-like
            Matrices of the SVD decomposition
        """
        if base_point is None:
            base_point = self.mean_estimator.fit(X).estimate_

        tangent_vecs = self._geometry.log(X, base_point=base_point)

        if self.space.point_ndim > 1:
            if gs.all(Matrices.is_square(tangent_vecs)) and gs.all(
                Matrices.is_symmetric(tangent_vecs)
            ):
                X = SymmetricMatrices.basis_representation(tangent_vecs)
            else:
                X = gs.reshape(tangent_vecs, (len(X), -1))
        else:
            X = tangent_vecs

        if self.n_components is None:
            n_components = min(X.shape)
        else:
            n_components = self.n_components

        n_samples, n_features = X.shape

        if n_components == "mle":
            if n_samples < n_features:
                raise ValueError(
                    "n_components='mle' is only supported if n_samples >= n_features"
                )
        elif not 0 <= n_components <= min(n_samples, n_features):
            raise ValueError(
                f"n_components={n_components} must be between 0 and "
                f"min(n_samples, n_features)={min(n_samples, n_features)} with "
                "svd_solver='full'"
            )
        elif n_components >= 1 and not isinstance(n_components, numbers.Integral):
            raise ValueError(
                f"n_components={n_components} must be of type int "
                "when greater than or equal to 1, "
                f"was of type={type(n_components)}"
            )

        # Center data - the mean should be 0 if base_point is the Frechet mean
        self.mean_ = gs.mean(X, axis=0)
        X -= self.mean_

        U, S, V = gs.linalg.svd(X, full_matrices=False)
        # flip eigenvectors' sign to enforce deterministic output
        U, V = svd_flip(U, V)

        components_ = V

        # Get variance explained by singular values
        explained_variance_ = (S**2) / (n_samples - 1)
        total_var = explained_variance_.sum()
        explained_variance_ratio_ = explained_variance_ / total_var
        singular_values_ = gs.copy(S)  # Store the singular values.

        # Postprocess the number of components required
        if n_components == "mle":
            n_components = _infer_dimension_(explained_variance_, n_samples, n_features)
        elif 0 < n_components < 1.0:
            # number of components for which the cumulated explained
            # variance percentage is superior to the desired threshold
            ratio_cumsum = stable_cumsum(explained_variance_ratio_)
            n_components = gs.searchsorted(ratio_cumsum, n_components) + 1

        # Compute noise covariance using Probabilistic PCA model
        # The sigma2 maximum likelihood (cf. eq. 12.46)
        if n_components < min(n_features, n_samples):
            self.noise_variance_ = explained_variance_[n_components:].mean()
        else:
            self.noise_variance_ = 0.0

        self.base_point_ = base_point
        self.n_samples_, self.n_features_ = n_samples, n_features
        self.components_ = components_[:n_components]
        self.n_components_ = int(n_components)
        self.explained_variance_ = explained_variance_[:n_components]
        self.explained_variance_ratio_ = explained_variance_ratio_[:n_components]
        self.singular_values_ = singular_values_[:n_components]

        return U, S, V


class HyperbolicPlaneExactPGA(_BasePCA):
    """Exact Principal Geodesic Analysis in the hyperbolic plane.

    The first principal component is computed by finding the direction
    in a unit ball around the mean that maximizes the variance of the
    projections on the induced geodesic. The projections are given by
    closed form expressions in extrinsic coordinates. The second principal
    component is the direction at the mean that is orthogonal to the first
    principal component.

    Parameters
    ----------
    space : Hyperbolic
        Two-dimensional hyperbolic space.
    n_vec : int
        Number of vectors used to discretize the unit ball when finding
        the direction of maximal variance.

    Attributes
    ----------
    components_ : array-like, shape=[n_components, 2]
        Principal axes, representing the directions of maximal variance in the
        data. They are the initial velocities of the principal geodesics.
    mean_ : array-like, shape=[2,]
        Intrinsic mean of the data points.

    References
    ----------
    .. [CSV2016] R. Chakraborty, D. Seo, and B. C. Vemuri,
        "An efficient exact-pga algorithm for constant curvature manifolds."
        Proceedings of the IEEE conference on computer vision and pattern
        recognition. 2016.
    """

    def __init__(self, space, n_grid=100):
        self.space = space
        if self.space.dim != 2:
            raise NotImplementedError(
                "Exact PGA is only implemented for the 2-dimensional hyperbolic space."
            )
        self.n_grid = n_grid
        self.mean_estimator = FrechetMean(space=self.space)
        self._half_space = Hyperbolic(2, coords_type="half-space")
        self._space_ext = Hyperbolic(2, coords_type="extrinsic")

    def _variance_of_projections(self, pt_ext, mn_ext, vec_ext):
        projections = self._space_ext.project_on_geodesic(pt_ext, mn_ext, vec_ext)
        costs = self._space_ext.metric.dist(mn_ext, projections) ** 2
        return gs.sum(costs)

    def fit(self, X, y=None):
        """Fit the model with X.

        Parameters
        ----------
        X : array-like, shape=[..., n_features]
            Training data in the hyperbolic plane. If the space is
            the Poincare half-space or Poincare ball, n_features is
            2. If it is the hyperboloid, n_features is 3.
        y : Ignored (Compliance with scikit-learn interface)

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        self.mean_ = self.mean_estimator.fit(X).estimate_

        mean_half_space = self.space.to_coordinates(self.mean_, "half-space")
        mean_ext = self.space.to_coordinates(self.mean_, "extrinsic")
        X_ext = self.space.to_coordinates(X, "extrinsic")

        angles_half_space = gs.linspace(0.0, 2 * gs.pi, self.n_grid)
        angles_half_space = gs.expand_dims(angles_half_space, axis=1)
        vectors_half_space = gs.hstack(
            (gs.cos(angles_half_space), gs.sin(angles_half_space))
        )
        norms = self._half_space.metric.norm(vectors_half_space, mean_half_space)
        vectors_half_space = gs.einsum("ij,i->ij", vectors_half_space, 1 / norms)
        vectors_ext = self.space.half_space_to_extrinsic_tangent(
            vectors_half_space, mean_half_space
        )

        costs = gs.array(
            [
                self._variance_of_projections(X_ext, mean_ext, vec_ext)
                for vec_ext in vectors_ext
            ]
        )
        axis_1 = vectors_half_space[gs.argmax(costs)]
        axis_2 = gs.array([-axis_1[1], axis_1[0]])
        components_half_space = gs.stack((axis_1, axis_2))
        self.components_ = self.space.from_tangent_coordinates(
            components_half_space, mean_half_space, "half-space"
        )
        return self

    def fit_transform(self, X, y=None):
        """Project X on the principal components.

        Parameters
        ----------
         X : array-like, shape=[n_points, 2]
             Training data in the hyperbolic plane. If the space is
             the Poincare half-space or Poincare ball, n_features is
             2. If it is the hyperboloid, n_features is 3.
         y : Ignored (Compliance with scikit-learn interface)

        Returns
        -------
        X_new : array-like, shape=[n_components, n_points, 2]
            Projections of the data on the first principal geodesic (first line
            of the array) and on the second principal geodesic (second line).
        """
        self.fit(X)
        axis_1, axis_2 = self.components_
        axis_1_ext = self.space.to_tangent_coordinates(axis_1, self.mean_, "extrinsic")
        axis_2_ext = self.space.to_tangent_coordinates(axis_2, self.mean_, "extrinsic")
        X_ext = self.space.to_coordinates(X, "extrinsic")
        mean_ext = self.space.to_coordinates(self.mean_, "extrinsic")

        proj1_ext = self._space_ext.project_on_geodesic(X_ext, mean_ext, axis_1_ext)
        proj2_ext = self._space_ext.project_on_geodesic(X_ext, mean_ext, axis_2_ext)

        var_1 = gs.mean(self._space_ext.metric.dist(mean_ext, proj1_ext) ** 2)
        var_2 = gs.mean(self._space_ext.metric.dist(mean_ext, proj2_ext) ** 2)
        self.explained_variance_ = gs.stack((var_1, var_2))

        return gs.stack(
            [
                self.space.from_coordinates(proj1_ext, "extrinsic"),
                self.space.from_coordinates(proj2_ext, "extrinsic"),
            ]
        )


class ExactPGA:
    r"""Exact Principal Geodesic Analysis.

    Parameters
    ----------
    space : Manifold
        Equipped manifold.
    """

    def __new__(cls, space, **kwargs):
        """Interface for instantiating proper algorithm."""
        if isinstance(space, _Hyperbolic) and space.dim == 2:
            return HyperbolicPlaneExactPGA(space, **kwargs)

        else:
            raise NotImplementedError(
                "Exact PGA is only implemented for the two-dimensional "
                "hyperbolic space."
            )



from ._dispatch import BACKEND_NAME, _common
from ._dispatch import numpy as _np

_is_iterable = _common._is_iterable
_is_boolean = _common._is_boolean
_get_wider_dtype = _common._get_wider_dtype
array = _common.array
cast = _common.cast
convert_to_wider_dtype = _common.convert_to_wider_dtype
eye = _common.eye
is_array = _common.is_array
get_default_dtype = _common.get_default_dtype
zeros = _common.zeros
_box_unary_scalar = _common._box_unary_scalar
_box_binary_scalar = _common._box_binary_scalar

abs = _box_unary_scalar(target=_np.abs)
arccos = _box_unary_scalar(target=_np.arccos)
arccosh = _box_unary_scalar(target=_np.arccosh)
arcsin = _box_unary_scalar(target=_np.arcsin)
arctanh = _box_unary_scalar(target=_np.arctanh)
ceil = _box_unary_scalar(target=_np.ceil)
cos = _box_unary_scalar(target=_np.cos)
cosh = _box_unary_scalar(target=_np.cosh)
exp = _box_unary_scalar(target=_np.exp)
floor = _box_unary_scalar(target=_np.floor)
log = _box_unary_scalar(target=_np.log)
sign = _box_unary_scalar(target=_np.sign)
sin = _box_unary_scalar(target=_np.sin)
sinh = _box_unary_scalar(target=_np.sinh)
sqrt = _box_unary_scalar(target=_np.sqrt)
tan = _box_unary_scalar(target=_np.tan)
tanh = _box_unary_scalar(target=_np.tanh)

arctan2 = _box_binary_scalar(target=_np.arctan2)
mod = _box_binary_scalar(target=_np.mod)
power = _box_binary_scalar(target=_np.power)


def angle(z, deg=False):
    out = _np.angle(z, deg=deg)
    if isinstance(z, float):
        return cast(out, get_default_dtype())

    return out


def imag(x):
    out = _np.imag(x)
    if is_array(x):
        return out

    return get_default_dtype().type(out)


def real(x):
    out = _np.real(x)
    if is_array(x):
        return out

    return array(out)


def arange(start_or_stop, /, stop=None, step=1, dtype=None, **kwargs):
    if dtype is None and (
        isinstance(stop, float)
        or isinstance(step, float)
        or isinstance(start_or_stop, float)
    ):
        dtype = get_default_dtype()

    if stop is None:
        return _np.arange(start_or_stop, step=step, dtype=dtype)

    return _np.arange(start_or_stop, stop, step=step, dtype=dtype)


def to_numpy(x):
    return x


def from_numpy(x):
    return x


def squeeze(x, axis=None):
    if axis is None:
        return _np.squeeze(x)
    if x.shape[axis] != 1:
        return x
    return _np.squeeze(x, axis=axis)


def flatten(x):
    return x.flatten()


def one_hot(labels, num_classes):
    return eye(num_classes, dtype=_np.dtype("uint8"))[labels]


def assignment(x, values, indices, axis=0):
    """Assign values at given indices of an array.

    Parameters
    ----------
    x: array-like, shape=[dim]
        Initial array.
    values: {float, list(float)}
        Value or list of values to be assigned.
    indices: {int, tuple, list(int), list(tuple)}
        Single int or tuple, or list of ints or tuples of indices where value
        is assigned.
        If the length of the tuples is shorter than ndim(x), values are
        assigned to each copy along axis.
    axis: int, optional
        Axis along which values are assigned, if vectorized.

    Returns
    -------
    x_new : array-like, shape=[dim]
        Copy of x with the values assigned at the given indices.

    Notes
    -----
    If a single value is provided, it is assigned at all the indices.
    If a list is given, it must have the same length as indices.
    """
    x_new = copy(x)

    use_vectorization = hasattr(indices, "__len__") and len(indices) < ndim(x)
    if _is_boolean(indices):
        x_new[indices] = values
        return x_new
    zip_indices = _is_iterable(indices) and _is_iterable(indices[0])
    len_indices = len(indices) if _is_iterable(indices) else 1
    if zip_indices:
        indices = tuple(zip(*indices))
    if not use_vectorization:
        if not zip_indices:
            len_indices = len(indices) if _is_iterable(indices) else 1
        len_values = len(values) if _is_iterable(values) else 1
        if len_values > 1 and len_values != len_indices:
            raise ValueError("Either one value or as many values as indices")
        x_new[indices] = values
    else:
        indices = tuple(list(indices[:axis]) + [slice(None)] + list(indices[axis:]))
        x_new[indices] = values
    return x_new


def assignment_by_sum(x, values, indices, axis=0):
    """Add values at given indices of an array.

    Parameters
    ----------
    x : array-like, shape=[dim]
        Initial array.
    values : {float, list(float)}
        Value or list of values to be assigned.
    indices : {int, tuple, list(int), list(tuple)}
        Single int or tuple, or list of ints or tuples of indices where value
        is assigned.
        If the length of the tuples is shorter than ndim(x), values are
        assigned to each copy along axis.
    axis: int, optional
        Axis along which values are assigned, if vectorized.

    Returns
    -------
    x_new : array-like, shape=[dim]
        Copy of x with the values assigned at the given indices.

    Notes
    -----
    If a single value is provided, it is assigned at all the indices.
    If a list is given, it must have the same length as indices.
    """
    x_new = copy(x)

    use_vectorization = hasattr(indices, "__len__") and len(indices) < ndim(x)
    if _is_boolean(indices):
        x_new[indices] += values
        return x_new
    zip_indices = _is_iterable(indices) and _is_iterable(indices[0])
    if zip_indices:
        indices = tuple(zip(*indices))
    if not use_vectorization:
        len_indices = len(indices) if _is_iterable(indices) else 1
        len_values = len(values) if _is_iterable(values) else 1
        if len_values > 1 and len_values != len_indices:
            raise ValueError("Either one value or as many values as indices")
        x_new[indices] += values
    else:
        indices = tuple(list(indices[:axis]) + [slice(None)] + list(indices[axis:]))
        x_new[indices] += values
    return x_new


def ndim(x):
    return x.ndim


def get_slice(x, indices):
    """Return a slice of an array, following Numpy's style.

    Parameters
    ----------
    x : array-like, shape=[dim]
        Initial array.
    indices : iterable(iterable(int))
        Indices which are kept along each axis, starting from 0.

    Returns
    -------
    slice : array-like
        Slice of x given by indices.

    Notes
    -----
    This follows Numpy's convention: indices are grouped by axis.

    Examples
    --------
    >>> a = np.array(range(30)).reshape(3,10)
    >>> get_slice(a, ((0, 2), (8, 9)))
    array([8, 29])
    """
    return x[indices]


def vectorize(x, pyfunc, multiple_args=False, signature=None, **kwargs):
    if multiple_args:
        return _np.vectorize(pyfunc, signature=signature)(*x)
    return _np.vectorize(pyfunc, signature=signature)(x)


def set_diag(x, new_diag):
    """Set the diagonal along the last two axis.

    Parameters
    ----------
    x : array-like, shape=[dim]
        Initial array.
    new_diag : array-like, shape=[dim[-2]]
        Values to set on the diagonal.

    Returns
    -------
    None

    Notes
    -----
    This mimics tensorflow.linalg.set_diag(x, new_diag), when new_diag is a
    1-D array, but modifies x instead of creating a copy.
    """
    arr_shape = x.shape
    x[..., range(arr_shape[-2]), range(arr_shape[-1])] = new_diag
    return x


def copy(x):
    return x.copy()


def array_from_sparse(indices, data, target_shape):
    """Create an array of given shape, with values at specific indices.

    The rest of the array will be filled with zeros.

    Parameters
    ----------
    indices : iterable(tuple(int))
        Index of each element which will be assigned a specific value.
    data : iterable(scalar)
        Value associated at each index.
    target_shape : tuple(int)
        Shape of the output array.

    Returns
    -------
    a : array, shape=target_shape
        Array of zeros with specified values assigned to specified indices.
    """
    data = array(data)
    out = zeros(target_shape, dtype=data.dtype)
    out.put(_np.ravel_multi_index(_np.array(indices).T, target_shape), data)
    return out


def vec_to_diag(vec):
    """Convert vector to diagonal matrix."""
    d = vec.shape[-1]
    return _np.squeeze(vec[..., None, :] * eye(d, dtype=vec.dtype)[None, :, :])


def tril_to_vec(x, k=0):
    n = x.shape[-1]
    rows, cols = _np.tril_indices(n, k=k)
    return x[..., rows, cols]


def triu_to_vec(x, k=0):
    n = x.shape[-1]
    rows, cols = _np.triu_indices(n, k=k)
    return x[..., rows, cols]


def mat_from_diag_triu_tril(diag, tri_upp, tri_low):
    """Build matrix from given components.

    Forms a matrix from diagonal, strictly upper triangular and
    strictly lower traingular parts.

    Parameters
    ----------
    diag : array_like, shape=[..., n]
    tri_upp : array_like, shape=[..., (n * (n - 1)) / 2]
    tri_low : array_like, shape=[..., (n * (n - 1)) / 2]

    Returns
    -------
    mat : array_like, shape=[..., n, n]
    """
    diag, tri_upp, tri_low = convert_to_wider_dtype([diag, tri_upp, tri_low])

    n = diag.shape[-1]
    (i,) = _np.diag_indices(n, ndim=1)
    j, k = _np.triu_indices(n, k=1)
    mat = zeros(diag.shape + (n,), dtype=diag.dtype)
    mat[..., i, i] = diag
    mat[..., j, k] = tri_upp
    mat[..., k, j] = tri_low
    return mat


def divide(a, b, ignore_div_zero=False):
    if ignore_div_zero is False:
        return _np.divide(a, b)

    wider_dtype, _ = _get_wider_dtype([a, b])
    return _np.divide(a, b, out=zeros(a.shape, dtype=wider_dtype), where=b != 0)


def ravel_tril_indices(n, k=0, m=None):
    if m is None:
        size = (n, n)
    else:
        size = (n, m)
    idxs = _np.tril_indices(n, k, m)
    return _np.ravel_multi_index(idxs, size)


def matmul(*args, **kwargs):
    for arg in args:
        if arg.ndim == 1:
            raise ValueError("ndims must be >=2")
    return _np.matmul(*args, **kwargs)


def outer(a, b):
    if a.ndim > 1 and b.ndim > 1:
        return _np.einsum("...i,...j->...ij", a, b)

    out = _np.multiply.outer(a, b)
    if b.ndim > 1:
        out = out.swapaxes(0, -2)

    return out


def matvec(A, b):
    if b.ndim == 1:
        return _np.matmul(A, b)
    if A.ndim == 2:
        return _np.matmul(A, b.T).T
    return _np.einsum("...ij,...j->...i", A, b)


def dot(a, b):
    if b.ndim == 1:
        return _np.dot(a, b)

    if a.ndim == 1:
        return _np.dot(a, b.T)

    return _np.einsum("...i,...i->...", a, b)


def trace(a):
    return _np.trace(a, axis1=-2, axis2=-1)


def scatter_add(input, dim, index, src):
    """Add values from src into input at the indices specified in index.

    Parameters
    ----------
    input : array-like
        Tensor to scatter values into.
    dim : int
        The axis along which to index.
    index : array-like
        The indices of elements to scatter.
    src : array-like
        The source element(s) to scatter.

    Returns
    -------
    input : array-like
        Modified input array.
    """
    if dim == 0:
        for i, val in zip(index, src):
            input[i] += val
        return input
    if dim == 1:
        for j in range(len(input)):
            for i, val in zip(index[j], src[j]):
                if not isinstance(val, _np.float64) and BACKEND_NAME == "autograd":
                    val = float(val._value)
                input[j, i] += float(val)
        return input
    raise NotImplementedError



from ._dispatch import _common
from ._dispatch import numpy as _np
from ._dispatch import scipy as _scipy

_to_ndarray = _common.to_ndarray
_cast_fout_to_input_dtype = _common._cast_fout_to_input_dtype
_cast_out_to_input_dtype = _common._cast_out_to_input_dtype
atol = _common.atol


def _is_symmetric(x, tol=atol):
    new_x = _to_ndarray(x, to_ndim=3)
    return (_np.abs(new_x - _np.transpose(new_x, axes=(0, 2, 1))) < tol).all()


def _is_hermitian(x, tol=atol):
    new_x = _to_ndarray(x, to_ndim=3)
    return (_np.abs(new_x - _np.conj(_np.transpose(new_x, axes=(0, 2, 1)))) < tol).all()


_diag_vec = _np.vectorize(_np.diag, signature="(n)->(n,n)")

_logm_vec = _cast_fout_to_input_dtype(
    target=_np.vectorize(_scipy.linalg.logm, signature="(n,m)->(n,m)")
)


def logm(x):
    ndim = x.ndim
    new_x = _to_ndarray(x, to_ndim=3)

    if _is_symmetric(new_x) and new_x.dtype not in [_np.complex64, _np.complex128]:
        eigvals, eigvecs = _np.linalg.eigh(new_x)
        if (eigvals > 0).all():
            eigvals = _np.log(eigvals)
            eigvals = _diag_vec(eigvals)
            transp_eigvecs = _np.transpose(eigvecs, axes=(0, 2, 1))
            result = _np.matmul(eigvecs, eigvals)
            result = _np.matmul(result, transp_eigvecs)
        else:
            result = _logm_vec(new_x)
    else:
        result = _logm_vec(new_x)

    if ndim == 2:
        return result[0]
    return result


def solve_sylvester(a, b, q, tol=atol):
    if a.shape == b.shape:
        axes = (0, 2, 1) if a.ndim == 3 else (1, 0)
        if _np.all(_np.isclose(a, b)) and _np.all(
            _np.abs(a - _np.transpose(a, axes)) < tol
        ):
            eigvals, eigvecs = _np.linalg.eigh(a)
            if _np.all(eigvals >= tol):
                tilde_q = _np.transpose(eigvecs, axes) @ q @ eigvecs
                tilde_x = tilde_q / (eigvals[..., :, None] + eigvals[..., None, :])
                return eigvecs @ tilde_x @ _np.transpose(eigvecs, axes)

    return _np.vectorize(
        _scipy.linalg.solve_sylvester, signature="(m,m),(n,n),(m,n)->(m,n)"
    )(a, b, q)


@_cast_fout_to_input_dtype
def sqrtm(x):
    return _np.vectorize(_scipy.linalg.sqrtm, signature="(n,m)->(n,m)")(x)


def quadratic_assignment(a, b, options):
    return list(_scipy.optimize.quadratic_assignment(a, b, options=options).col_ind)


def qr(*args, **kwargs):
    return _np.vectorize(
        _np.linalg.qr, signature="(n,m)->(n,k),(k,m)", excluded=["mode"]
    )(*args, **kwargs)


def is_single_matrix_pd(mat):
    """Check if 2D square matrix is positive definite."""
    if mat.shape[0] != mat.shape[1]:
        return False
    if mat.dtype in [_np.complex64, _np.complex128]:
        if not _is_hermitian(mat):
            return False
        eigvals = _np.linalg.eigvalsh(mat)
        return _np.min(_np.real(eigvals)) > 0
    try:
        _np.linalg.cholesky(mat)
        return True
    except _np.linalg.LinAlgError as e:
        if e.args[0] == "Matrix is not positive definite":
            return False
        raise e


@_cast_out_to_input_dtype
def fractional_matrix_power(A, t):
    if A.ndim == 2:
        return _scipy.linalg.fractional_matrix_power(A, t)

    return _np.stack([_scipy.linalg.fractional_matrix_power(A_, t) for A_ in A])


def polar(*args, **kwargs):
    """Polar decomposition of a matrix."""
    return _np.vectorize(
        _scipy.linalg.polar, signature="(n,n)->(n,n),(n,n)", excluded=["side"]
    )(*args, **kwargs)


def solve(a, b):
    """
    Solve a linear matrix equation, or system of linear scalar equations.

    Computes the "exact" solution, `x`, of the well-determined, i.e., full
    rank, linear matrix equation `ax = b`.

    Parameters
    ----------
    a : array-like, shape=[..., M, M]
        Coefficient matrix.
    b : array-like, shape=[..., M]
        Ordinate or "dependent variable" values".

    Returns
    -------
    x : array-like, shape=[..., M]
        Solution to the system a x = b.
    """
    batch_shape = a.shape[:-2]
    if batch_shape:
        b = _np.expand_dims(b, axis=-1)

    res = _np.linalg.solve(a, b)
    if batch_shape:
        return res[..., 0]

    return res



import pytest

import geomstats.backend as gs
from geomstats.geometry.matrices import Matrices
from geomstats.test.test_case import TestCase
from geomstats.test.vectorization import generate_vectorization_data
from geomstats.test_cases.geometry.base import MatrixVectorSpaceTestCase
from geomstats.test_cases.geometry.euclidean import EuclideanMetricTestCase


class MatrixOperationsTestCase(TestCase):
    @staticmethod
    def _get_transpose_shape(mat):
        return mat.shape[::-1][:2]

    def test_equal(self, mat_a, mat_b, expected, atol):
        res = Matrices.equal(mat_a, mat_b, atol=atol)
        self.assertAllEqual(res, expected)

    @pytest.mark.vec
    def test_equal_vec(self, n_reps, atol):
        mat_a = self.data_generator.random_mat()
        mat_b = self.data_generator.random_mat(shape=mat_a.shape[-2:])

        expected = Matrices.equal(mat_a, mat_b)

        vec_data = generate_vectorization_data(
            data=[dict(mat_a=mat_a, mat_b=mat_b, expected=expected, atol=atol)],
            arg_names=["mat_a", "mat_b"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_equal_true(self, n_points, atol):
        mat_a = mat_b = self.data_generator.random_mat(n_points)

        self.test_equal(mat_a, mat_b, True, atol=atol)

    @pytest.mark.random
    def test_equal_false(self, n_points, atol):
        mat_a = self.data_generator.random_mat(n_points)
        mat_b = self.data_generator.random_mat(n_points, shape=mat_a.shape[-2:])

        self.test_equal(mat_a, mat_b, False, atol=atol)

    def test_mul_reduce(self, mats, expected, atol):
        res = Matrices.mul(*mats)
        self.assertAllClose(res, expected, atol=atol)

    def test_mul(self, mat_a, mat_b, expected, atol):
        res = Matrices.mul(mat_a, mat_b)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_mul_vec(self, n_reps, atol):
        mat_a = self.data_generator.random_mat()
        transpose_shape = self._get_transpose_shape(mat_a)
        mat_b = self.data_generator.random_mat(shape=transpose_shape)

        expected = Matrices.mul(mat_a, mat_b)

        vec_data = generate_vectorization_data(
            data=[dict(mat_a=mat_a, mat_b=mat_b, expected=expected, atol=atol)],
            arg_names=["mat_a", "mat_b"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_mul_identity(self, n_points, atol):
        mat_a = self.data_generator.random_mat(n_points)
        identity = gs.eye(mat_a.shape[-1])

        self.test_mul(mat_a, identity, mat_a, atol)

    def test_bracket(self, mat_a, mat_b, expected, atol):
        res = Matrices.bracket(mat_a, mat_b)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_bracket_vec(self, n_reps, atol):
        mat_a = self.data_generator.random_square_mat()
        mat_b = self.data_generator.random_square_mat(shape=mat_a.shape[-2:])

        expected = Matrices.bracket(mat_a, mat_b)

        vec_data = generate_vectorization_data(
            data=[dict(mat_a=mat_a, mat_b=mat_b, expected=expected, atol=atol)],
            arg_names=["mat_a", "mat_b"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_bracket_bilinearity(self, n_points, atol):
        mat_a = self.data_generator.random_square_mat(n_points)
        shape = mat_a.shape[-2:]
        mat_b = self.data_generator.random_square_mat(shape=shape)
        mat_c = self.data_generator.random_square_mat(shape=shape)

        a, b = gs.random.uniform(size=(2,))

        mat_ab = a * mat_a + b * mat_b
        left = Matrices.bracket(mat_ab, mat_c)
        right = a * Matrices.bracket(mat_a, mat_c) + b * Matrices.bracket(mat_b, mat_c)
        self.assertAllClose(left, right, atol=atol)

        left = Matrices.bracket(mat_c, mat_ab)
        right = a * Matrices.bracket(mat_c, mat_a) + b * Matrices.bracket(mat_c, mat_b)
        self.assertAllClose(left, right, atol=atol)

    def test_transpose(self, mat, expected, atol):
        res = Matrices.transpose(mat)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_transpose_vec(self, n_reps, atol):
        mat = self.data_generator.random_mat()
        expected = Matrices.transpose(mat)

        vec_data = generate_vectorization_data(
            data=[dict(mat=mat, expected=expected, atol=atol)],
            arg_names=["mat"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_transpose_sym(self, n_points, atol):
        mat = self.data_generator.random_symmetric_mat(n_points)

        self.test_transpose(mat, mat, atol)

    def test_diagonal(self, mat, expected, atol):
        res = Matrices.diagonal(mat)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_diagonal_vec(self, n_reps, atol):
        mat = self.data_generator.random_mat()

        expected = Matrices.diagonal(mat)
        vec_data = generate_vectorization_data(
            data=[dict(mat=mat, expected=expected, atol=atol)],
            arg_names=["mat"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_diagonal_sum(self, n_points, atol):
        mat = self.data_generator.random_square_mat(n_points)

        diag = Matrices.diagonal(mat)

        sum_diag = gs.sum(diag, axis=-1)
        trace = gs.trace(mat)
        self.assertAllClose(sum_diag, trace, atol=atol)

    def test_is_square(self, mat, expected):
        res = Matrices.is_square(mat)
        self.assertAllEqual(res, expected)

    @pytest.mark.vec
    def test_is_square_vec(self, n_reps):
        mat = self.data_generator.random_square_mat()

        expected = Matrices.is_square(mat)
        vec_data = generate_vectorization_data(
            data=[dict(mat=mat, expected=expected)],
            arg_names=["mat"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_is_square_true(self, n_points):
        mat = self.data_generator.random_square_mat(n_points)
        self.test_is_square(mat, True)

    @pytest.mark.random
    def test_is_square_false(self, n_points):
        mat = self.data_generator.random_non_square_mat(n_points)
        self.test_is_square(mat, False)

    def test_is_property(self, property_name, mat, expected, atol):
        func = getattr(Matrices, f"is_{property_name}")
        res = func(mat, atol=atol)
        self.assertAllEqual(res, expected)

    @pytest.mark.vec
    def test_is_property_vec(self, property_name, n_reps, atol):
        mat = getattr(self.data_generator, f"random_{property_name}_mat")()

        func = getattr(Matrices, f"is_{property_name}")
        expected = func(mat, atol=atol)

        vec_data = generate_vectorization_data(
            data=[
                dict(property_name=property_name, mat=mat, expected=expected, atol=atol)
            ],
            arg_names=["mat"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_is_property_true(self, property_name, n_points, atol):
        mat = getattr(self.data_generator, f"random_{property_name}_mat")(n_points)
        self.test_is_property(property_name, mat, True, atol)

    @pytest.mark.random
    def test_is_property_false_square(self, property_name, n_points, atol):
        mat = self.data_generator.random_square_mat(n_points)
        self.test_is_property(property_name, mat, False, atol)

    @pytest.mark.random
    def test_to_property_is_property(self, property_name, n_points, atol):
        square_mat = self.data_generator.random_square_mat(n_points)

        to_func = getattr(Matrices, f"to_{property_name}")
        is_func = getattr(Matrices, f"is_{property_name}")

        mat = to_func(square_mat)
        res = is_func(mat, atol=atol)

        self.assertAllEqual(res, True)

    @pytest.mark.random
    def test_to_lower_triangular_diagonal_scaled_is_lower_triangular(
        self, n_points, atol
    ):
        square_mat = self.data_generator.random_square_mat(n_points)

        mat = Matrices.to_lower_triangular_diagonal_scaled(square_mat)
        res = Matrices.is_lower_triangular(mat, atol=atol)

        self.assertAllEqual(res, True)

    def test_congruent(self, mat_1, mat_2, expected, atol):
        res = Matrices.congruent(
            mat_1,
            mat_2,
        )
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_congruent_vec(self, n_reps, atol):
        mat_1 = self.data_generator.random_square_mat()
        shape = mat_1.shape[-2:]
        mat_2 = self.data_generator.random_square_mat(shape=shape)

        expected = Matrices.congruent(mat_1, mat_2)

        vec_data = generate_vectorization_data(
            data=[dict(mat_1=mat_1, mat_2=mat_2, expected=expected, atol=atol)],
            arg_names=["mat_1", "mat_2"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    def test_frobenius_product(self, mat_1, mat_2, expected, atol):
        res = Matrices.frobenius_product(
            mat_1,
            mat_2,
        )
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_frobenius_product_vec(self, n_reps, atol):
        mat_1 = self.data_generator.random_mat()
        shape = mat_1.shape[-2:]
        mat_2 = self.data_generator.random_mat(shape=shape)

        expected = Matrices.frobenius_product(mat_1, mat_2)

        vec_data = generate_vectorization_data(
            data=[dict(mat_1=mat_1, mat_2=mat_2, expected=expected, atol=atol)],
            arg_names=["mat_1", "mat_2"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    def test_trace_product(self, mat_1, mat_2, expected, atol):
        res = Matrices.trace_product(
            mat_1,
            mat_2,
        )
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_trace_product_vec(self, n_reps, atol):
        mat_1 = self.data_generator.random_mat()
        shape = mat_1.shape[-2:]
        mat_2 = gs.transpose(self.data_generator.random_mat(shape=shape))

        expected = Matrices.trace_product(mat_1, mat_2)

        vec_data = generate_vectorization_data(
            data=[dict(mat_1=mat_1, mat_2=mat_2, expected=expected, atol=atol)],
            arg_names=["mat_1", "mat_2"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    def test_flatten(self, mat, expected, atol):
        res = Matrices.flatten(mat)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_flatten_vec(self, n_reps, atol):
        mat = self.data_generator.random_mat()

        expected = Matrices.flatten(mat)

        vec_data = generate_vectorization_data(
            data=[dict(mat=mat, expected=expected, atol=atol)],
            arg_names=["mat"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    def test_align_matrices(self, point, base_point, expected, atol):
        res = Matrices.align_matrices(point, base_point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_align_matrices_vec(self, n_reps, atol):
        point = self.data_generator.random_mat()
        shape = point.shape[-2:]
        base_point = self.data_generator.random_mat(shape=shape)

        expected = Matrices.align_matrices(point, base_point)

        vec_data = generate_vectorization_data(
            data=[
                dict(point=point, base_point=base_point, expected=expected, atol=atol)
            ],
            arg_names=["point", "base_point"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)


class MatricesTestCase(MatrixVectorSpaceTestCase):
    @pytest.mark.random
    def test_reshape_after_flatten(self, n_points, atol):
        point = self.data_generator.random_point(n_points)

        vec = Matrices.flatten(point)
        point_ = self.space.reshape(vec)
        self.assertAllClose(point, point_, atol=atol)


class MatricesMetricTestCase(EuclideanMetricTestCase):
    pass



import pytest

from geomstats.test.test_case import TestCase


class InterpolatorTestCase(TestCase):
    def test_interpolate(self, t, expected, atol):
        res = self.interpolator(t)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.random
    def test_interpolate_with_given_data(self, atol):
        res = self.interpolator(self.times)
        self.assertAllClose(res, self.data, atol=atol)

    @pytest.mark.random
    def test_interpolate_half_interval(self, atol):
        times = self.times[:-1] + 0.5 * (self.times[1:] - self.times[:-1])

        res = self.interpolator(times)

        point_ndim_slc = (slice(None),) * self.interpolator.point_ndim
        expected = self.data[(..., slice(0, -1)) + point_ndim_slc] + 0.5 * (
            self.data[(..., slice(1, None)) + point_ndim_slc]
            - self.data[(..., slice(0, -1)) + point_ndim_slc]
        )
        self.assertAllClose(res, expected, atol=atol)



"""The 3D Heisenberg group.

Lead author: Morten Pedersen.
"""

import geomstats.backend as gs
from geomstats.geometry.base import VectorSpace
from geomstats.geometry.euclidean import Euclidean
from geomstats.geometry.lie_group import LieGroup
from geomstats.geometry.symmetric_matrices import SymmetricMatrices


class HeisenbergVectors(LieGroup, VectorSpace):
    """Class for the 3D Heisenberg group in the vector representation.

    The 3D Heisenberg group represented as R^3. It is a step-2 Carnot Lie
    group. It can be equipped with a natural sub-Riemannian structure, and it
    is a fundamental example in sub-Riemannian geometry.

    Parameters
    ----------
    No parameters

    References
    ----------
    https://en.wikipedia.org/wiki/Heisenberg_group
    """

    def __init__(self, equip=True):
        super().__init__(dim=3, shape=(3,), lie_algebra=Euclidean(3), equip=equip)

    def _create_basis(self):
        """Create the canonical basis."""
        return gs.eye(3)

    @property
    def identity(self):
        """Identity of the 3D Heisenberg group.

        Returns
        -------
        _ : array-like, shape=[3,]
        """
        return gs.zeros(self.dim)

    def compose(self, point_a, point_b):
        """Compute the group product of elements `point_a` and `point_b`.

        Parameters
        ----------
        point_a : array-like, shape=[..., 3]
            Left factor in the product.
        point_b : array-like, shape=[..., 3]
            Right factor in the product.

        Returns
        -------
        point_ab : array-like, shape=[..., 3]
            Product of point_a and point_b along the first dimension.
        """
        point_ab = point_a + point_b
        point_ab_additional_term = gs.array(
            1
            / 2
            * (point_a[..., 0] * point_b[..., 1] - point_a[..., 1] * point_b[..., 0])
        )

        point_ab = point_ab + gs.concatenate(
            [gs.zeros_like(point_ab[..., :2]), point_ab_additional_term[..., None]],
            axis=-1,
        )

        return point_ab

    def inverse(self, point):
        """Compute the group inverse of point.

        Parameters
        ----------
        point : array-like, shape=[..., 3]
            Point.

        Returns
        -------
        _ : array-like, shape=[..., 3]
            Inverse.
        """
        return -point

    def jacobian_translation(self, point, left=True):
        """Compute the Jacobian matrix of left/right translation by a point.

        This calculates the differential of the left translation L_(point)
        evaluated at `point`. Note that it only depends on the point we are
        left-translating by, not on the point where the differential is
        evaluated.

        Parameters
        ----------
        point : array-like, shape=[..., 3]
            Point.
        left : bool
            Indicate whether to calculate the differential of left or right
            translations.
            Optional, default: True.

        Returns
        -------
        _ : array-like, shape=[..., 3, 3]
            Jacobian of the left/right translation by point.
        """
        e31 = gs.array_from_sparse([(2, 0)], [1.0], (3, 3))
        e32 = gs.array_from_sparse([(2, 1)], [1.0], (3, 3))

        if left:
            return (
                gs.eye(3)
                + gs.einsum("...,ij->...ij", -point[..., 1] / 2, e31)
                + gs.einsum("...,ij->...ij", point[..., 0] / 2, e32)
            )

        return (
            gs.eye(3)
            + gs.einsum("...,ij->...ij", point[..., 1] / 2, e31)
            + gs.einsum("...,ij->...ij", -point[..., 0] / 2, e32)
        )

    def exp_from_identity(self, tangent_vec):
        """Compute the group exponential of the tangent vector at the identity.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., 3]
            Tangent vector at base point.

        Returns
        -------
        _ : array-like, shape=[..., 3]
            Point.
        """
        return gs.copy(tangent_vec)

    def log_from_identity(self, point):
        """Compute the group logarithm of the point at the identity.

        Parameters
        ----------
        point : array-like, shape=[..., 3]
            Point.

        Returns
        -------
        _ : array-like, shape=[..., 3]
            Group logarithm.
        """
        return gs.copy(point)

    @staticmethod
    def upper_triangular_matrix_from_vector(point):
        """Compute the upper triangular matrix representation of the vector.

        The 3D Heisenberg group can also be represented as 3x3 upper triangular
        matrices. This function computes this representation of the vector
        `point`.

        Parameters
        ----------
        point : array-like, shape=[..., 3]
            Point in the vector-represention.

        Returns
        -------
        upper_triangular_mat : array-like, shape=[..., 3, 3]
            Upper triangular matrix.
        """
        element_02 = point[..., 2] + 1 / 2 * point[..., 0] * point[..., 1]

        if gs.ndim(point) == 1:
            modified_point = gs.array([1, point[0], element_02, 1, point[1], 1])
        else:
            n_points = gs.shape(point)[0]
            modified_point = gs.stack(
                (
                    gs.ones(n_points),
                    point[..., 0],
                    element_02,
                    gs.ones(n_points),
                    point[..., 1],
                    gs.ones(n_points),
                ),
                axis=-1,
            )

        return gs.triu(SymmetricMatrices.matrix_representation(modified_point))

    @staticmethod
    def vector_from_upper_triangular_matrix(matrix):
        """Compute the vector representation of the upper triangular matrix.

        Parameters
        ----------
        matrix : array-like, shape=[..., 3, 3]
            Upper triangular matrix.

        Returns
        -------
        vector : array-like, shape=[..., 3]
        """
        modified_point = gs.triu_to_vec(matrix, k=1)
        corrected_elem = (
            modified_point[..., 1]
            - 1 / 2 * modified_point[..., 0] * modified_point[..., 2]
        )
        return gs.stack(
            [modified_point[..., 0], modified_point[..., 2], corrected_elem], axis=-1
        )

    def lie_bracket(self, tangent_vec_a, tangent_vec_b, base_point=None):
        """Compute the lie bracket of two tangent vectors."""
        raise NotImplementedError("The lie bracket is not implemented.")



"""Implementation of sub-Riemannian geometry.

Lead author: Morten Pedersen.
"""

import geomstats.backend as gs


class SubRiemannianMetric:
    """Class for Sub-Riemannian metrics.

    This implementation assumes a bracket-generating distribution of constant dimension.

    Only one of the arguments 'cometric_matrix' and 'frame' can be different from
    None. If the frame is supplied, it is assumed orthonormal.

    Parameters
    ----------
    space : Manifold object
    cometric_matrix : callable
        Optional, default: 'None'

        The cometric matrix as a function of a point.

        You should pass :
            base_point : array-like, shape=[..., dim]
        It returns:
            _ : array-like, shape=[..., dim, dim]

    frame : callable
        Optional, default: 'None'

        Matrix representing the frame spanning the distribution,
        as a function of a point.

        You should pass:
            point : array-like, shape=[..., dim]
        It returns
            _ : array-like, shape=[..., dim, dist_dim]
                Frame field matrix. Each column is a vector field of the frame
                spanning the distribution.
    """

    def __init__(
        self,
        space,
        cometric_matrix=None,
        frame=None,
    ):
        if not bool(cometric_matrix is not None) ^ bool(frame is not None):
            raise ValueError(
                "Either 'cometric_matrix' or 'frame' must be passed, and not both."
            )

        if space.point_ndim != 1:
            raise ValueError("Can only handle 'vector' point types.")

        self.cometric_matrix = cometric_matrix
        self.frame = frame

        self._def_type = "cometric" if self.frame is None else "frame"

    def sr_sharp(self, base_point, cotangent_vec):
        r"""Compute sub-Riemannian sharp map.

        This is the sub-Riemannian sharp map, mapping a covector at base_point to a
        tangent vector in the distribution subspace at base_point. For an orthonormal
        frame (F_i)_{i=1..dist_dim}, the sharp map is given by

        .. math::
                \sharp(q, p) = \sum_i^{dist_dim} p(F_i(q)) * F_i(q)

        Parameters
        ----------
        base_point : array-like, shape=[..., dim]
            Point on the manifold.

        cotangent_vec : array-like, shape=[..., dim]
            Cotangent vector at `base_point`.

        Returns
        -------
        sr_sharp : array-like, shape=[..., dim]
            sub-Riemannian sharp of 'cotangent_vec' at 'base_point'
        """
        if self.frame is None:
            raise NotImplementedError(
                "The sub-Riemannian sharp map is only"
                " implemented when a frame is passed."
            )

        frame = self.frame(base_point)
        coefs = gs.einsum("...i,...ij->...j", cotangent_vec, frame)
        coefs_on_frame = gs.einsum("...j,...ij->...ij", coefs, frame)

        return gs.einsum("...ij->...i", coefs_on_frame)

    def _get_method(self, method_name):
        return getattr(self, f"_{method_name}_{self._def_type}")

    def _inner_coproduct_frame(self, cotangent_vec_a, cotangent_vec_b, base_point):
        sharp = self.sr_sharp(base_point=base_point, cotangent_vec=cotangent_vec_b)
        return gs.einsum("...i,...i -> ...", cotangent_vec_a, sharp)

    def _inner_coproduct_cometric(self, cotangent_vec_a, cotangent_vec_b, base_point):
        C_b = gs.einsum(
            "...ij,...j->...i", self.cometric_matrix(base_point), cotangent_vec_b
        )
        return gs.einsum("...i,...i->...", cotangent_vec_a, C_b)

    def inner_coproduct(self, cotangent_vec_a, cotangent_vec_b, base_point):
        """Compute inner coproduct between two cotangent vectors at base point.

        This is the inner product associated to the cometric matrix.

        Parameters
        ----------
        cotangent_vec_a : array-like, shape=[..., dim]
            Cotangent vector at `base_point`.
        cotangent_vec_b : array-like, shape=[..., dim]
            Cotangent vector at `base_point`.
        base_point : array-like, shape=[..., dim]
            Point on the manifold.

        Returns
        -------
        inner_coproduct : float
            Inner coproduct between the two cotangent vectors.
        """
        return self._get_method("inner_coproduct")(
            cotangent_vec_a, cotangent_vec_b, base_point
        )

    def _hamiltonian_frame(self, state):
        position, momentum = state

        inner_products = gs.einsum("...i,...ij->...j", momentum, self.frame(position))

        out = 0.5 * gs.einsum("...i,...i->...", inner_products, inner_products)
        if not gs.is_array(out):
            return gs.array(out)
        return out

    def _hamiltonian_cometric(self, state):
        position, momentum = state
        return 0.5 * self.inner_coproduct(momentum, momentum, position)

    def hamiltonian(self, state):
        r"""Compute the hamiltonian energy associated to the cometric.

        The Hamiltonian at state :math:`(q, p)` is defined by

        .. math::
                H(q, p) = \frac{1}{2} <p, p>_q

        where :math:`<\cdot, \cdot>_q` is the cometric at :math:`q`.

        Parameters
        ----------
        state : array-like, shape=[[..., dim], [..., dim]]
            The first array in 'state' contains positions, the second contains
            covectors (momentums).

        Returns
        -------
        energy :  float
            Hamiltonian energy at `state`.
        """
        return self._get_method("hamiltonian")(state)

    @staticmethod
    def symp_grad(hamiltonian):
        r"""Compute the symplectic gradient of the Hamiltonian.

        Parameters
        ----------
        hamiltonian : callable
            The hamiltonian function from the tangent bundle to the reals.

        Returns
        -------
        vector : array-like, shape=[, 2*dim]
            The symplectic gradient of the Hamiltonian.
        """
        value_and_grad = gs.autodiff.value_and_grad(
            hamiltonian,
            point_ndims=2,
        )

        def vector(x):
            """Compute symplectic gradient at x."""
            _, (h_q, h_p) = value_and_grad(x)
            return gs.array([h_p, -h_q])

        return vector

    def _symp_euler_frame(self, hamiltonian, step_size):
        def step(state):
            position, momentum = state
            dq = self.sr_sharp(base_point=position, cotangent_vec=momentum)
            y = gs.array([position + step_size * dq, momentum])
            _, dp = self.symp_grad(hamiltonian)(y)
            return gs.array([position + step_size * dq, momentum + step_size * dp])

        return step

    def _symp_euler_cometric(self, hamiltonian, step_size):
        def step(state):
            r"""Compute an integration step from state."""
            position, momentum = state
            dq, _ = self.symp_grad(hamiltonian)(state)
            y = gs.array([position + step_size * dq, momentum])
            _, dp = self.symp_grad(hamiltonian)(y)
            return gs.array([position + step_size * dq, momentum + step_size * dp])

        return step

    def symp_euler(self, hamiltonian, step_size):
        r"""Compute a function which calculates a step of symplectic euler integration.

        Parameters
        ----------
        hamiltonian : callable
            The hamiltonian function from the tangent bundle to the reals.
        step_size : float
            Step size of the symplectic euler step

        Returns
        -------
        step : callable
            Given a state, 'step' returns the next symplectic euler step
        """
        return self._get_method("symp_euler")(hamiltonian, step_size)

    @staticmethod
    def iterate(func, n_steps):
        r"""Construct a function which iterates another function n_steps times.

        Parameters
        ----------
        func : callable
            A function which calculates the next step of a
            sequence to be calculated.
        n_steps : int
            The number of times to iterate func.

        Returns
        -------
        flow : callable
            Given a state, 'flow' returns a sequence with n_steps
            iterations of func.
        """

        def flow(x):
            r"""Compute flow starting at x."""
            xs = [x]
            for i in range(n_steps):
                xs.append(func(xs[i]))
            return gs.array(xs)

        return flow

    def symp_flow(self, hamiltonian, end_time=1.0, n_steps=20):
        r"""Compute the symplectic flow of the hamiltonian.

        Parameters
        ----------
        hamiltonian : callable
            The hamiltonian function from the tangent bundle to the reals.
        end_time : float
            The last time point until which to calculate the flow.
        n_steps : int
            Number of integration steps.

        Returns
        -------
        _ : array-like, shape[,n_steps]
            Given a state, 'symp_flow' returns a
            sequence with n_steps iterations of func.
        """
        step = self.symp_euler
        step_size = end_time / n_steps
        return self.iterate(step(hamiltonian, step_size), n_steps)

    def exp(self, cotangent_vec, base_point, n_steps=20):
        """Exponential map associated to the cometric.

        Exponential map at base_point of cotangent_vec computed by integration
        of the Hamiltonian equation (initial value problem), using the cometric.
        In the Riemannian case this yields the exponential associated to the
        Levi-Civita connection of the metric (the inverse of the cometric).

        Parameters
        ----------
        cotangent_vec : array-like, shape=[..., dim]
            Tangent vector at the base point.
        base_point : array-like, shape=[..., dim]
            Point on the manifold.
        n_steps : int
            Number of discrete time steps to take in the integration.
            Optional, default: N_STEPS.

        Returns
        -------
        exp : array-like, shape=[..., dim]
            Point on the manifold.
        """
        # TODO: add log (caveat in docstrings)
        # TODO: add distance (based on log - also caveat)
        base_point = gs.broadcast_to(base_point, cotangent_vec.shape)

        # TODO: integrate; allow euler
        initial_state = gs.stack([base_point, cotangent_vec])

        flow = self.symp_flow(self.hamiltonian, n_steps=n_steps)

        return flow(initial_state)[-1][0]

    def geodesic(self, initial_point, initial_cotangent_vec, n_steps=20):
        """Generate parameterized function for the normal geodesic curve.

        Normal geodesic curve defined by an initial point and an initial
        cotangent vector.

        Parameters
        ----------
        initial_point : array-like, shape=[..., dim]
            Point on the manifold, initial point of the geodesic.
        initial_cotangent_vec : array-like, shape=[..., dim]
            Cotangent vector at base point, the initial speed of the geodesics.

        Returns
        -------
        path : callable
            Time parameterized normal geodesic curve. If a batch of initial
            conditions is passed, the output array's first dimension
            represents the different initial conditions, and the second
            corresponds to time.
        """
        if initial_cotangent_vec is None:
            raise ValueError(
                "Specify an initial cotangent " "vector to define the geodesic."
            )

        initial_point = gs.to_ndarray(initial_point, to_ndim=2)
        initial_cotangent_vec = gs.to_ndarray(initial_cotangent_vec, to_ndim=2)

        n_initial_conditions = initial_cotangent_vec.shape[0]

        if n_initial_conditions > 1 and len(initial_point) == 1:
            initial_point = gs.stack([initial_point[0]] * n_initial_conditions)

        def path(t):
            """Generate parameterized function for geodesic curve.

            Parameters
            ----------
            t : array-like, shape=[n_points,]
                Times at which to compute points of the geodesics.
            """
            t = gs.array(t)
            t = gs.cast(t, initial_cotangent_vec.dtype)
            t = gs.to_ndarray(t, to_ndim=1)

            cotangent_vecs = gs.einsum("i,...k->...ik", t, initial_cotangent_vec)

            points_at_time_t = [
                self.exp(tv, pt, n_steps=n_steps)
                for tv, pt in zip(cotangent_vecs, initial_point)
            ]
            points_at_time_t = gs.stack(points_at_time_t, axis=0)

            return (
                points_at_time_t[0] if n_initial_conditions == 1 else points_at_time_t
            )

        return path



import pytest

import geomstats.backend as gs
from geomstats.test_cases.geometry.base import VectorSpaceOpenSetTestCase
from geomstats.test_cases.geometry.riemannian_metric import RiemannianMetricTestCase
from geomstats.test_cases.information_geometry.base import (
    InformationManifoldMixinTestCase,
)


class DirichletDistributionsTestCase(
    InformationManifoldMixinTestCase, VectorSpaceOpenSetTestCase
):
    def _check_sample_belongs_to_support(self, sample, atol):
        self.assertTrue(gs.all(sample >= 0.0))
        self.assertTrue(gs.all(gs.abs(gs.sum(sample, axis=-1) - 1.0) < atol))


class DirichletMetricTestCase(RiemannianMetricTestCase):
    def test_jacobian_christoffels(self, base_point, expected, atol):
        res = self.space.metric.jacobian_christoffels(base_point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.random
    def test_sectional_curvature_is_negative(self, n_points):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec_a = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_b = self.data_generator.random_tangent_vec(base_point)

        res = self.space.metric.sectional_curvature(
            tangent_vec_a, tangent_vec_b, base_point
        )
        self.assertTrue(gs.all(res <= 0))

    @pytest.mark.random
    def test_exp_diagonal_is_totally_geodesic(self, param, tangent_param, atol):
        """Check that the diagonal x1 = ... = xn is totally geodesic."""
        base_point = param * gs.ones(self.space.shape)
        tangent_vec = tangent_param * gs.ones(self.space.shape)

        end_point = self.space.metric.exp(tangent_vec, base_point)
        self.assertAllClose(
            end_point - end_point[0], gs.zeros(self.space.shape), atol=atol
        )



"""Visualization for Geometric Statistics."""

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D  # NOQA

import geomstats.backend as gs
from geomstats.geometry.hypersphere import Hypersphere

S1 = Hypersphere(dim=1)
S2 = Hypersphere(dim=2)

AX_SCALE = 1.2


class Circle:
    """Class used to draw a circle."""

    def __init__(self, n_angles=100, points=None):
        angles = gs.linspace(0.0, 2 * gs.pi, n_angles + 1)
        self.circle_x = gs.cos(angles)
        self.circle_y = gs.sin(angles)
        self.points = []
        if points is not None:
            self.add_points(points)

    @staticmethod
    def set_ax(ax=None):
        """Set axis."""
        if ax is None:
            ax = plt.subplot()
        ax_s = AX_SCALE
        plt.setp(
            ax,
            xlim=(-ax_s, ax_s),
            ylim=(-ax_s, ax_s),
            xlabel="X",
            ylabel="Y",
        )
        return ax

    def add_points(self, points):
        """Add points."""
        if not gs.all(S1.belongs(points)):
            raise ValueError("Points do  not belong to the circle.")
        if not isinstance(points, list):
            points = list(points)
        self.points.extend(points)

    def draw(self, ax, **plot_kwargs):
        """Plot circle shape."""
        ax.plot(self.circle_x, self.circle_y, color="black")
        if self.points:
            self.draw_points(ax, **plot_kwargs)

    def draw_points(self, ax, points=None, **plot_kwargs):
        """Plot points."""
        if points is None:
            points = self.points
        points = gs.array(points)
        ax.plot(points[:, 0], points[:, 1], marker="o", linestyle="None", **plot_kwargs)

    def plot(self, points, ax=None, **point_draw_kwargs):
        """Plot points in the circle."""
        ax = self.set_ax(ax=ax)
        self.add_points(points)
        self.draw(ax, **point_draw_kwargs)


class Sphere:
    """Create the arrays sphere_x, sphere_y, sphere_z to plot a sphere.

    Create the arrays sphere_x, sphere_y, sphere_z of values
    to plot the wireframe of a sphere.
    Their shape is (n_meridians, n_circles_latitude).
    """

    def __init__(self, n_meridians=40, n_circles_latitude=None, points=None):
        if n_circles_latitude is None:
            n_circles_latitude = max(n_meridians // 2, 4)

        u, v = gs.meshgrid(
            gs.linspace(0.0, 2 * gs.pi, n_meridians + 1),
            gs.linspace(0.0, gs.pi, n_circles_latitude + 1),
        )

        self.center = gs.zeros(3)
        self.radius = 1
        self.sphere_x = self.center[0] + self.radius * gs.cos(u) * gs.sin(v)
        self.sphere_y = self.center[1] + self.radius * gs.sin(u) * gs.sin(v)
        self.sphere_z = self.center[2] + self.radius * gs.cos(v)

        self.points = []
        if points is not None:
            self.add_points(points)

    @staticmethod
    def set_ax(ax=None):
        """Set axis."""
        if ax is None:
            ax = plt.subplot(111, projection="3d")

        ax_s = AX_SCALE
        plt.setp(
            ax,
            xlim=(-ax_s, ax_s),
            ylim=(-ax_s, ax_s),
            zlim=(-ax_s, ax_s),
            xlabel="X",
            ylabel="Y",
            zlabel="Z",
        )
        ax.set_box_aspect([1.0, 1.0, 1.0])
        return ax

    def add_points(self, points):
        """Add points."""
        if not gs.all(S2.belongs(points)):
            raise ValueError("Points do not belong to the sphere.")
        if not isinstance(points, list):
            points = list(points)
        self.points.extend(points)

    def draw(self, ax, **scatter_kwargs):
        """Plot sphere shape."""
        ax.plot_wireframe(
            self.sphere_x, self.sphere_y, self.sphere_z, color="grey", alpha=0.2
        )
        ax.set_box_aspect([1.0, 1.0, 1.0])
        if self.points:
            self.draw_points(ax, **scatter_kwargs)

    def draw_points(self, ax, points=None, **scatter_kwargs):
        """Plot points."""
        if points is None:
            points = self.points
        points = [gs.to_numpy(point) for point in points]
        points_x = [point[0] for point in points]
        points_y = [point[1] for point in points]
        points_z = [point[2] for point in points]
        ax.scatter(points_x, points_y, points_z, **scatter_kwargs)

        for i_point, point in enumerate(points):
            if "label" in scatter_kwargs:
                if len(scatter_kwargs["label"]) == len(points):
                    ax.text(
                        point[0],
                        point[1],
                        point[2],
                        scatter_kwargs["label"][i_point],
                        size=10,
                        zorder=1,
                        color="k",
                    )

    def get_fibonnaci_points(self, n_points=16000):
        """Get spherical Fibonacci point sets.

        Point sets are yield nearly uniform point distributions on the unit
        sphere.
        """
        x_vals = []
        y_vals = []
        z_vals = []

        offset = 2.0 / n_points
        increment = gs.pi * (3.0 - gs.sqrt(5.0))

        for i in range(n_points):
            y = ((i * offset) - 1) + (offset / 2)
            r = gs.sqrt(1 - pow(y, 2))

            phi = ((i + 1) % n_points) * increment

            x = gs.cos(phi) * r
            z = gs.sin(phi) * r

            x_vals.append(x)
            y_vals.append(y)
            z_vals.append(z)

        x_vals = [(self.radius * i) for i in x_vals]
        y_vals = [(self.radius * i) for i in y_vals]
        z_vals = [(self.radius * i) for i in z_vals]

        return gs.array([x_vals, y_vals, z_vals])

    def plot_heatmap(self, ax, scalar_function, n_points=16000, alpha=0.2, cmap="jet"):
        """Plot a heatmap defined by a loss on the sphere."""
        points = self.get_fibonnaci_points(n_points)
        intensity = gs.array([scalar_function(x) for x in points.T])
        ax.scatter(
            points[0, :],
            points[1, :],
            points[2, :],
            c=intensity,
            alpha=alpha,
            marker=".",
            cmap=plt.get_cmap(cmap),
        )

    def plot(self, points, ax=None, **point_draw_kwargs):
        """Plot points in the sphere."""
        ax = self.set_ax(ax=ax)
        self.points = []
        self.add_points(points)
        self.draw(ax, **point_draw_kwargs)



"""Incremental frechet mean estimator."""

from sklearn.base import BaseEstimator


class IncrementalFrechetMean(BaseEstimator):
    r"""Incremental Frechet Mean Estimator.

    Incremental frechet mean estimator calculates sample frechet mean by
    moving iteratively along the geodesic between current mean estimate
    and next point.

    .. math::
        \text{Initialization}: m_{1} := X_{1}

    .. math::
        \text{Update}: \text{Let } \gamma_k \text{ be geodesic joining }
        m_{k-1}\text{ and } X_{k} \text{ then }
        m_{k} := \gamma(1/k) \,\, \forall  2 \leq k \leq N

    Asymptotic convergence to population frechet mean is guranteed for
    simply connected, complete and non-positively curved Riemannian manifolds.
    It is important to note that estimator obtained by such iterative fashion
    need not necessarily be solution to the following optimization problem.

    .. math::
        \max_{q \in M} \sum_{i=1}^{N} d(q, X_{i})^2

    where d is the riemannian metric. Also, Estimator is not permutation
    invariant , i.e.,the estimate might depend on the order in which
    incremental updates are performed.

    Parameters
    ----------
    space : Manifold
        Equipped manifold.
    verbose : bool
        Verbose option.
        Optional, default: False.
    clean_state : bool
        If keeping track of last iteration or clean state of estimator.

    Notes
    -----
    * Required metric methods: `geodesic`.

    References
    ----------
    .. [CHSV2016] Cheng, Ho, Salehian, Vemuri.
        "Recursive Computation of the Frechet Mean on Non-Positively
        Curved Riemannian Manifolds with Applications",
        Riemannian Computing in Computer Vision pp 21-43, 2016.
        https://link.springer.com/chapter/10.1007/978-3-319-22957-7_2
    """

    def __init__(
        self,
        space,
        verbose=False,
        clean_state=True,
    ):
        self.space = space
        self.verbose = verbose
        self.clean_state = clean_state

        self.iter = 0
        self.estimate_ = None

    def fit(self, X, y=None, init=None):
        """Compute the incremental Frechet mean.

        Parameters
        ----------
        X : array-like, shape=[n_samples, {dim, [n, n]}]
            Training input samples.
        y : None
            Ignored.
        init : array-like, shape=[{dim, [n, n]}]
            If not None, starts mean computation from init, could be useful
            when data comes in streaming setting.
            Optional, default: None.

        Returns
        -------
        self : object
            Returns self.
        """
        N = X.shape[0]
        if init is not None:
            m_curr = init
            idxs = range(N)
        else:
            m_curr = X[0]
            self.iter += 1
            idxs = range(1, N)

        for i in idxs:
            geod_func = self.space.metric.geodesic(m_curr, X[i])
            m_curr = geod_func(1 / (self.iter + 1))[0]
            self.iter += 1

        if self.clean_state:
            self.iter = 0

        self.estimate_ = m_curr
        return self



"""The manifold of symmetric positive definite (SPD) matrices.

Lead authors: Yann Thanwerdas and Olivier Bisson.
"""

import math

import geomstats.backend as gs
from geomstats.geometry.base import VectorSpaceOpenSet
from geomstats.geometry.complex_matrices import ComplexMatrices
from geomstats.geometry.diffeo import Diffeo
from geomstats.geometry.general_linear import GeneralLinear
from geomstats.geometry.hermitian_matrices import apply_func_to_eigvalsh, expmh, powermh
from geomstats.geometry.matrices import Matrices, MatricesMetric
from geomstats.geometry.positive_lower_triangular_matrices import (
    InvariantPositiveLowerTriangularMatricesMetric,
    PositiveLowerTriangularMatrices,
)
from geomstats.geometry.pullback_metric import PullbackDiffeoMetric
from geomstats.geometry.riemannian_metric import RiemannianMetric
from geomstats.geometry.scalar_product_metric import ScalarProductMetric
from geomstats.geometry.symmetric_matrices import SymmetricMatrices
from geomstats.integrator import integrate
from geomstats.vectorization import repeat_out


def logmh(mat):
    """Compute the matrix log for a Hermitian matrix."""
    n = mat.shape[-1]
    dim_3_mat = gs.reshape(mat, [-1, n, n])
    logm = apply_func_to_eigvalsh(dim_3_mat, gs.log, check_positive=True)
    return gs.reshape(logm, mat.shape)


def _aux_differential_power(power, tangent_vec, base_point):
    """Compute the differential of the matrix power.

    Auxiliary function to the functions differential_power and
    inverse_differential_power.

    Parameters
    ----------
    power : float
        Power function to differentiate.
    tangent_vec : array_like, shape=[..., n, n]
        Tangent vector at base point.
    base_point : array_like, shape=[..., n, n]
        Base point.

    Returns
    -------
    eigvectors : array-like, shape=[..., n, n]
    transp_eigvectors : array-like, shape=[..., n, n]
    numerator : array-like, shape=[..., n, n]
    denominator : array-like, shape=[..., n, n]
    temp_result : array-like, shape=[..., n, n]
    """
    eigvalues, eigvectors = gs.linalg.eigh(base_point)

    if power == 0:
        powered_eigvalues = gs.log(eigvalues)
    elif power == math.inf:
        powered_eigvalues = gs.exp(eigvalues)
    else:
        powered_eigvalues = eigvalues**power

    denominator = eigvalues[..., :, None] - eigvalues[..., None, :]
    numerator = powered_eigvalues[..., :, None] - powered_eigvalues[..., None, :]

    null_denominator = gs.abs(denominator) < gs.atol
    if power == 0:
        numerator = gs.where(null_denominator, gs.ones_like(numerator), numerator)
        denominator = gs.where(null_denominator, eigvalues[..., :, None], denominator)
    elif power == math.inf:
        numerator = gs.where(
            null_denominator, powered_eigvalues[..., :, None], numerator
        )
        denominator = gs.where(null_denominator, gs.ones_like(numerator), denominator)
    else:
        numerator = gs.where(
            null_denominator, power * powered_eigvalues[..., :, None], numerator
        )
        denominator = gs.where(null_denominator, eigvalues[..., :, None], denominator)

    if gs.is_complex(base_point):
        transp_eigvectors = ComplexMatrices.transconjugate(eigvectors)
    else:
        transp_eigvectors = Matrices.transpose(eigvectors)

    temp_result = Matrices.mul(transp_eigvectors, tangent_vec, eigvectors)

    if gs.is_complex(base_point):
        numerator = gs.cast(numerator, dtype=temp_result.dtype)
        denominator = gs.cast(denominator, dtype=temp_result.dtype)

    return (eigvectors, transp_eigvectors, numerator, denominator, temp_result)


class SymMatrixLog(Diffeo):
    """Matrix logarithm diffeomorphism.

    A diffeomorphism from the space of symmetric positive-definite matrices to
    the space of symmetric matrices.
    """

    @classmethod
    def __call__(cls, base_point):
        """Compute the matrix log for a symmetric matrix.

        Parameters
        ----------
        base_point : array_like, shape=[..., n, n]
            Symmetric matrix.

        Returns
        -------
        log : array_like, shape=[..., n, n]
            Matrix logarithm of base_point.
        """
        return logmh(base_point)

    @classmethod
    def inverse(cls, image_point):
        """Compute the matrix exponential for a symmetric matrix.

        Parameters
        ----------
        image_point : array_like, shape=[..., n, n]
            Symmetric matrix.

        Returns
        -------
        exponential : array_like, shape=[..., n, n]
            Exponential of image_point.
        """
        return expmh(image_point)

    @classmethod
    def tangent(cls, tangent_vec, base_point=None, image_point=None):
        """Compute the differential of the matrix logarithm.

        Compute the differential of the matrix logarithm on SPD
        matrices at base_point applied to tangent_vec.

        Parameters
        ----------
        tangent_vec : array_like, shape=[..., n, n]
            Tangent vector at base point.
        base_point : array_like, shape=[..., n, n]
            Base point.
        image_point : array_like, shape=[..., n, n]
            Image base point.

        Returns
        -------
        differential_log : array-like, shape=[..., n, n]
            Differential of the matrix logarithm.
        """
        if base_point is not None:
            (
                eigvectors,
                transp_eigvectors,
                numerator,
                denominator,
                temp_result,
            ) = _aux_differential_power(0, tangent_vec, base_point)
            power_operator = numerator / denominator
        else:
            (
                eigvectors,
                transp_eigvectors,
                numerator,
                denominator,
                temp_result,
            ) = _aux_differential_power(math.inf, tangent_vec, image_point)
            power_operator = denominator / numerator

        result = power_operator * temp_result
        return Matrices.mul(eigvectors, result, transp_eigvectors)

    @classmethod
    def inverse_tangent(cls, image_tangent_vec, image_point=None, base_point=None):
        """Compute the differential of the matrix exponential.

        Computes the differential of the matrix exponential on SPD
        matrices at base_point applied to tangent_vec.

        Parameters
        ----------
        image_tangent_vec : array_like, shape=[..., n, n]
            Image tangent vector at image point.
        image_point : array_like, shape=[..., n, n]
            Image point.
        base_point : array_like, shape=[..., n, n]
            Base point.

        Returns
        -------
        differential_exp : array-like, shape=[..., n, n]
            Differential of the matrix exponential.
        """
        if image_point is not None:
            (
                eigvectors,
                transp_eigvectors,
                numerator,
                denominator,
                temp_result,
            ) = _aux_differential_power(math.inf, image_tangent_vec, image_point)
            power_operator = numerator / denominator
        else:
            (
                eigvectors,
                transp_eigvectors,
                numerator,
                denominator,
                temp_result,
            ) = _aux_differential_power(0, image_tangent_vec, base_point)
            power_operator = denominator / numerator

        result = power_operator * temp_result
        return Matrices.mul(eigvectors, result, transp_eigvectors)


class MatrixPower(Diffeo):
    """Matrix power diffeomorphism.

    A diffeomorphism from the space of symmetric positive-definite matrices to
    itself.
    """

    def __init__(self, power):
        self.power = power

    def __call__(self, base_point):
        """Compute the matrix power.

        Parameters
        ----------
        base_point : array_like, shape=[..., n, n]
            Symmetric matrix with non-negative eigenvalues.

        Returns
        -------
        powerm : array_like or list of arrays, shape=[..., n, n]
            Matrix power of mat.
        """
        return powermh(base_point, self.power)

    def inverse(self, image_point):
        """Compute the inverse matrix power.

        Parameters
        ----------
        image_point : array_like, shape=[..., n, n]
            Symmetric matrix with non-negative eigenvalues.

        Returns
        -------
        powerm : array_like or list of arrays, shape=[..., n, n]
            Matrix power of mat.
        """
        return powermh(image_point, 1 / self.power)

    def tangent(self, tangent_vec, base_point=None, image_point=None):
        r"""Compute the differential of the matrix power function.

        Compute the differential of the power function on SPD(n),
        :math:`A^p=\exp(p \log(A))`, at base_point applied to tangent_vec.

        Parameters
        ----------
        tangent_vec : array_like, shape=[..., n, n]
            Tangent vector at base point.
        base_point : array_like, shape=[..., n, n]
            Base point.
        image_point : array_like, shape=[..., n, n]
            Image base point.

        Returns
        -------
        differential_power : array-like, shape=[..., n, n]
            Differential of the power function.
        """
        if base_point is None:
            base_point = self.inverse(image_point)
        (
            eigvectors,
            transp_eigvectors,
            numerator,
            denominator,
            temp_result,
        ) = _aux_differential_power(self.power, tangent_vec, base_point)

        power_operator = numerator / denominator
        result = power_operator * temp_result
        return Matrices.mul(eigvectors, result, transp_eigvectors)

    def inverse_tangent(self, image_tangent_vec, image_point=None, base_point=None):
        r"""Compute the inverse of the differential of the matrix power.

        Compute the inverse of the differential of the power
        function on SPD matrices, :math:`A^p=\exp(p \log(A))`, at base_point
        applied to tangent_vec.

        Parameters
        ----------
        image_tangent_vec : array_like, shape=[..., n, n]
            Image tangent vector at image point.
        image_base_point : array_like, shape=[..., n, n]
            Image point.
        base_point : array_like, shape=[..., n, n]
            Base point.

        Returns
        -------
        inverse_differential_power : array-like, shape=[..., n, n]
            Inverse of the differential of the power function.
        """
        if base_point is None:
            base_point = self.inverse(image_point)
        (
            eigvectors,
            transp_eigvectors,
            numerator,
            denominator,
            temp_result,
        ) = _aux_differential_power(self.power, image_tangent_vec, base_point)
        power_operator = denominator / numerator
        result = power_operator * temp_result
        return Matrices.mul(eigvectors, result, transp_eigvectors)


class CholeskyMap(Diffeo):
    """Cholesky map.

    A diffeomorphism from the space of symmetric positive-definite matrices to
    the space of positive lower triangular matrices.
    """

    @classmethod
    def __call__(cls, base_point):
        """Compute cholesky factor.

        Compute cholesky factor for a symmetric positive definite matrix.

        Parameters
        ----------
        base_point : array_like, shape=[..., n, n]
            spd matrix.

        Returns
        -------
        cf : array_like, shape=[..., n, n]
            lower triangular matrix with positive diagonal elements.
        """
        return gs.linalg.cholesky(base_point)

    @staticmethod
    def inverse(image_point):
        """Compute gram matrix of rows.

        Gram_matrix is mapping from point to point.point^{T}.
        This is diffeomorphism between cholesky space and spd manifold.

        Parameters
        ----------
        image_point : array-like, shape=[..., n, n]
            element in cholesky space.

        Returns
        -------
        projected: array-like, shape=[..., n, n]
            SPD matrix.
        """
        return gs.einsum("...ij,...kj->...ik", image_point, image_point)

    @classmethod
    def tangent(cls, tangent_vec, base_point=None, image_point=None):
        """Compute the differential of the cholesky factor map.

        Parameters
        ----------
        tangent_vec : array_like, shape=[..., n, n]
            Tangent vector at base point.
        base_point : array_like, shape=[..., n, n]
            Base point.
        image_point : array_like, shape=[..., n, n]
            Image base point.

        Returns
        -------
        differential_cf : array-like, shape=[..., n, n]
            Differential of cholesky factor map
            lower triangular matrix.
        """
        if image_point is None:
            image_point = cls.__call__(base_point)

        inv_base_point = gs.linalg.inv(image_point)
        inv_transpose_base_point = Matrices.transpose(inv_base_point)
        aux = Matrices.to_lower_triangular_diagonal_scaled(
            Matrices.mul(inv_base_point, tangent_vec, inv_transpose_base_point)
        )
        return Matrices.mul(image_point, aux)

    @classmethod
    def inverse_tangent(cls, image_tangent_vec, image_point=None, base_point=None):
        """Compute differential of gram.

        Parameters
        ----------
        image_tangent_vec : array_like, shape=[..., n, n]
            Tangent vector at base point.
        image_point : array_like, shape=[..., n, n]
            Base point.
        base_point : array_like, shape=[..., n, n]
            Base point.

        Returns
        -------
        differential_gram : array-like, shape=[..., n, n]
            Differential of the gram.
        """
        if image_point is None:
            image_point = cls.__call__(base_point)

        mat1 = gs.einsum("...ij,...kj->...ik", image_tangent_vec, image_point)
        mat2 = gs.einsum("...ij,...kj->...ik", image_point, image_tangent_vec)
        return mat1 + mat2


class SPDMatrices(VectorSpaceOpenSet):
    """Class for the manifold of symmetric positive definite (SPD) matrices.

    Parameters
    ----------
    n : int
        Integer representing the shape of the matrices: n x n.
    """

    def __init__(self, n, equip=True):
        super().__init__(
            dim=int(n * (n + 1) / 2),
            embedding_space=SymmetricMatrices(n),
            equip=equip,
        )
        self.n = n

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return SPDAffineMetric

    def belongs(self, point, atol=gs.atol):
        """Check if a matrix is symmetric with positive eigenvalues.

        Parameters
        ----------
        mat : array-like, shape=[..., n, n]
            Matrix to be checked.
        atol : float
            Tolerance.
            Optional, default: backend atol.

        Returns
        -------
        belongs : array-like, shape=[...,]
            Boolean denoting if mat is an SPD matrix.
        """
        is_sym = self.embedding_space.belongs(point, atol)
        is_pd = Matrices.is_pd(point)
        return gs.logical_and(is_sym, is_pd)

    def projection(self, point):
        """Project a matrix to the space of SPD matrices.

        First the symmetric part of point is computed, then the eigenvalues
        are floored to gs.atol.

        Parameters
        ----------
        point : array-like, shape=[..., n, n]
            Matrix to project.

        Returns
        -------
        projected: array-like, shape=[..., n, n]
            SPD matrix.
        """
        sym = Matrices.to_symmetric(point)
        eigvals, eigvecs = gs.linalg.eigh(sym)
        regularized = gs.where(eigvals < gs.atol, gs.atol, eigvals)
        reconstruction = gs.einsum("...ij,...j->...ij", eigvecs, regularized)
        return Matrices.mul(reconstruction, Matrices.transpose(eigvecs))

    def random_point(self, n_samples=1, bound=1.0):
        """Sample in SPD(n) from the log-uniform distribution.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.
        bound : float
            Bound of the interval in which to sample in the tangent space.
            Optional, default: 1.

        Returns
        -------
        samples : array-like, shape=[..., n, n]
            Points sampled in SPD(n).
        """
        n = self.n
        size = (n_samples, n, n) if n_samples != 1 else (n, n)

        mat = bound * (2 * gs.random.rand(*size) - 1)
        return GeneralLinear.exp(Matrices.to_symmetric(mat))

    def random_tangent_vec(self, base_point, n_samples=1):
        """Sample on the tangent space of SPD(n) from the uniform distribution.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.
        base_point : array-like, shape=[..., n, n]
            Base point of the tangent space.

        Returns
        -------
        samples : array-like, shape=[..., n, n]
            Points sampled in the tangent space at base_point.
        """
        n = self.n
        size = (n_samples, n, n) if n_samples != 1 else (n, n)

        sqrt_base_point = gs.linalg.sqrtm(base_point)

        tangent_vec_at_id_aux = 2 * gs.random.rand(*size) - 1
        tangent_vec_at_id = tangent_vec_at_id_aux + Matrices.transpose(
            tangent_vec_at_id_aux
        )

        return Matrices.mul(sqrt_base_point, tangent_vec_at_id, sqrt_base_point)


class SPDAffineMetric(RiemannianMetric):
    """Class for the affine-invariant metric on the SPD manifold.

    Parameters
    ----------
    power_affine : int
        Power transformation of the classical SPD metric.
        Optional, default: 1.

    References
    ----------
    .. [TP2019] Thanwerdas, Pennec. "Is affine-invariance well defined on
        SPD matrices? A principled continuum of metrics" Proc. of GSI,
        2019. https://arxiv.org/abs/1906.01349
    """

    def inner_product(self, tangent_vec_a, tangent_vec_b, base_point):
        """Compute the affine-invariant inner-product.

        Compute the inner-product of tangent_vec_a and tangent_vec_b
        at point base_point using the affine invariant Riemannian metric.

        Parameters
        ----------
        tangent_vec_a : array-like, shape=[..., n, n]
            Tangent vector at base point.
        tangent_vec_b : array-like, shape=[..., n, n]
            Tangent vector at base point.
        base_point : array-like, shape=[..., n, n]
            Base point.

        Returns
        -------
        inner_product : array-like, shape=[..., n, n]
            Inner-product.
        """
        inv_base_point = GeneralLinear.inverse(base_point)
        aux_a = Matrices.mul(inv_base_point, tangent_vec_a)
        aux_b = Matrices.mul(inv_base_point, tangent_vec_b)

        return Matrices.trace_product(aux_a, aux_b)

    def exp(self, tangent_vec, base_point):
        """Compute the affine-invariant exponential map.

        Compute the Riemannian exponential at point base_point
        of tangent vector tangent_vec wrt the metric defined in inner_product.
        This gives a symmetric positive definite matrix.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., n, n]
            Tangent vector at base point.
        base_point : array-like, shape=[..., n, n]
            Base point.

        Returns
        -------
        exp : array-like, shape=[..., n, n]
            Riemannian exponential.
        """
        sqrt_base_point, inv_sqrt_base_point = powermh(base_point, [1.0 / 2, -1.0 / 2])

        tangent_vec_at_id = Matrices.mul(
            inv_sqrt_base_point, tangent_vec, inv_sqrt_base_point
        )

        tangent_vec_at_id = Matrices.to_symmetric(tangent_vec_at_id)
        exp_from_id = expmh(tangent_vec_at_id)

        return Matrices.mul(sqrt_base_point, exp_from_id, sqrt_base_point)

    def log(self, point, base_point):
        """Compute the affine-invariant logarithm map.

        Compute the Riemannian logarithm at point base_point,
        of point wrt the metric defined in inner_product.
        This gives a tangent vector at point base_point.

        Parameters
        ----------
        point : array-like, shape=[..., n, n]
            Point.
        base_point : array-like, shape=[..., n, n]
            Base point.

        Returns
        -------
        log : array-like, shape=[..., n, n]
            Riemannian logarithm of point at base_point.
        """
        sqrt_base_point, inv_sqrt_base_point = powermh(base_point, [1.0 / 2, -1.0 / 2])
        point_near_id = Matrices.mul(inv_sqrt_base_point, point, inv_sqrt_base_point)
        point_near_id = Matrices.to_symmetric(point_near_id)

        log_at_id = logmh(point_near_id)
        return Matrices.mul(sqrt_base_point, log_at_id, sqrt_base_point)

    def parallel_transport(
        self, tangent_vec, base_point, direction=None, end_point=None
    ):
        r"""Parallel transport of a tangent vector.

        Closed-form solution for the parallel transport of a tangent vector
        along the geodesic between two points `base_point` and `end_point`
        or alternatively defined by :math:`t \mapsto exp_{(base\_point)}(
        t*direction)`.
        Denoting `tangent_vec_a` by `S`, `base_point` by `A`, and `end_point`
        by `B` or `B = Exp_A(tangent_vec_b)` and :math:`E = (BA^{- 1})^{( 1
        / 2)}`. Then the parallel transport to `B` is:

        .. math::
            S' = ESE^T

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., n, n]
            Tangent vector at base point to be transported.
        base_point : array-like, shape=[..., n, n]
            Point on the manifold of SPD matrices. Point to transport from
        direction : array-like, shape=[..., n, n]
            Tangent vector at base point, initial speed of the geodesic along
            which the parallel transport is computed. Unused if `end_point` is given.
            Optional, default: None.
        end_point : array-like, shape=[..., n, n]
            Point on the manifold of SPD matrices. Point to transport to.
            Optional, default: None.

        Returns
        -------
        transported_tangent_vec: array-like, shape=[..., n, n]
            Transported tangent vector at exp_(base_point)(tangent_vec_b).
        """
        if end_point is None:
            end_point = self.exp(direction, base_point)
        # compute B^1/2(B^-1/2 A B^-1/2)B^-1/2 instead of sqrtm(AB^-1)
        sqrt_bp, inv_sqrt_bp = powermh(base_point, [1.0 / 2, -1.0 / 2])
        pdt = powermh(Matrices.mul(inv_sqrt_bp, end_point, inv_sqrt_bp), 1.0 / 2)
        congruence_mat = Matrices.mul(sqrt_bp, pdt, inv_sqrt_bp)
        return Matrices.congruent(tangent_vec, congruence_mat)

    def injectivity_radius(self, base_point=None):
        """Radius of the largest ball where the exponential is injective.

        Because of the negative curvature of this space, the injectivity radius
        is infinite everywhere.

        Parameters
        ----------
        base_point : array-like, shape=[..., n, n]
            Point on the manifold.

        Returns
        -------
        radius : array-like, shape=[...,]
            Injectivity radius.
        """
        radius = gs.array(math.inf)
        return repeat_out(self._space.point_ndim, radius, base_point)


class SPDBuresWassersteinMetric(RiemannianMetric):
    """Class for the Bures-Wasserstein metric on the SPD manifold.

    References
    ----------
    .. [BJL2017] Bhatia, Jain, Lim. "On the Bures-Wasserstein distance between
        positive definite matrices" Elsevier, Expositiones Mathematicae,
        vol. 37(2), 165-191, 2017. https://arxiv.org/pdf/1712.01504.pdf
    .. [MMP2018] Malago, Montrucchio, Pistone. "Wasserstein-Riemannian
        geometry of Gaussian densities"  Information Geometry, vol. 1, 137-179,
        2018. https://arxiv.org/pdf/1801.09269.pdf
    """

    def inner_product(self, tangent_vec_a, tangent_vec_b, base_point):
        r"""Compute the Bures-Wasserstein inner-product.

        Compute the inner-product of tangent_vec_a :math:`A` and tangent_vec_b
        :math:`B` at point base_point :math:`S=PDP^\top` using the
        Bures-Wasserstein Riemannian metric:

        .. math::
            \frac{1}{2}\sum_{i,j}\frac{[P^\top AP]_{ij}[P^\top BP]_{ij}}{d_i+d_j}

        Parameters
        ----------
        tangent_vec_a : array-like, shape=[..., n, n]
            Tangent vector at base point.
        tangent_vec_b : array-like, shape=[..., n, n]
            Tangent vector at base point.
        base_point : array-like, shape=[..., n, n]
            Base point.

        Returns
        -------
        inner_product : array-like, shape=[...,]
            Inner-product.
        """
        eigvals, eigvecs = gs.linalg.eigh(base_point)
        transp_eigvecs = Matrices.transpose(eigvecs)
        rotated_tangent_vec_a = Matrices.mul(transp_eigvecs, tangent_vec_a, eigvecs)
        rotated_tangent_vec_b = Matrices.mul(transp_eigvecs, tangent_vec_b, eigvecs)

        coefficients = 1 / (eigvals[..., :, None] + eigvals[..., None, :])
        result = (
            Matrices.frobenius_product(
                coefficients * rotated_tangent_vec_a, rotated_tangent_vec_b
            )
            / 2
        )

        return result

    def exp(self, tangent_vec, base_point):
        """Compute the Bures-Wasserstein exponential map.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., n, n]
            Tangent vector at base point.
        base_point : array-like, shape=[..., n, n]
            Base point.

        Returns
        -------
        exp : array-like, shape=[...,]
            Riemannian exponential.
        """
        eigvals, eigvecs = gs.linalg.eigh(base_point)
        transp_eigvecs = Matrices.transpose(eigvecs)
        rotated_tangent_vec = Matrices.mul(transp_eigvecs, tangent_vec, eigvecs)
        coefficients = 1 / (eigvals[..., :, None] + eigvals[..., None, :])
        rotated_sylvester = rotated_tangent_vec * coefficients
        rotated_hessian = gs.einsum("...ij,...j->...ij", rotated_sylvester, eigvals)
        rotated_hessian = Matrices.mul(rotated_hessian, rotated_sylvester)
        hessian = Matrices.mul(eigvecs, rotated_hessian, transp_eigvecs)

        return base_point + tangent_vec + hessian

    def log(self, point, base_point):
        """Compute the Bures-Wasserstein logarithm map.

        Compute the Riemannian logarithm at point base_point,
        of point wrt the Bures-Wasserstein metric.
        This gives a tangent vector at point base_point.

        Parameters
        ----------
        point : array-like, shape=[..., n, n]
            Point.
        base_point : array-like, shape=[..., n, n]
            Base point.

        Returns
        -------
        log : array-like, shape=[..., n, n]
            Riemannian logarithm.
        """
        # compute B^1/2(B^-1/2 A B^-1/2)B^-1/2 instead of sqrtm(AB^-1)
        sqrt_bp, inv_sqrt_bp = powermh(base_point, [0.5, -0.5])
        pdt = powermh(Matrices.mul(sqrt_bp, point, sqrt_bp), 0.5)
        sqrt_product = Matrices.mul(sqrt_bp, pdt, inv_sqrt_bp)
        transp_sqrt_product = Matrices.transpose(sqrt_product)
        return sqrt_product + transp_sqrt_product - 2 * base_point

    def squared_dist(self, point_a, point_b):
        """Compute the Bures-Wasserstein squared distance.

        Compute the Riemannian squared distance between point_a and point_b.

        Parameters
        ----------
        point_a : array-like, shape=[..., n, n]
            Point.
        point_b : array-like, shape=[..., n, n]
            Point.

        Returns
        -------
        squared_dist : array-like, shape=[...]
            Riemannian squared distance.

        Notes
        -----
        Use of `abs` in the output prevents nan when calling
        `sqrt` in very small negative outputs (e.g. -1e-16).
        """
        La, Qa = gs.linalg.eigh(point_a)
        point_a_sqrt = Matrices.mul(
            Qa,
            gs.vec_to_diag(
                gs.sqrt(La * (La > 0)),
            ),
            Matrices.transpose(Qa),
        )

        Lc, Qc = gs.linalg.eigh(point_a_sqrt @ point_b @ point_a_sqrt)
        cross_term = Matrices.mul(
            Qc,
            gs.vec_to_diag(gs.sqrt(Lc * (Lc > 0))),
            Matrices.transpose(Qc),
        )

        return gs.abs(gs.trace(point_a + point_b - 2 * cross_term))

    def parallel_transport(
        self,
        tangent_vec,
        base_point,
        direction=None,
        end_point=None,
        n_steps=10,
        step="rk4",
    ):
        r"""Compute the parallel transport of a tangent vec along a geodesic.

        Approximation of the solution of the parallel transport of a tangent
        vector a along the geodesic defined by :math:`t \mapsto exp_{(
        base\_point)}(t* tangent\_vec\_b)`. The parallel transport equation is
        formulated in this case in [TP2021]_.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., n, n]
            Tangent vector at `base_point` to transport.
        base_point : array-like, shape=[..., n, n]
            Initial point of the geodesic.
        direction : array-like, shape=[..., n, n]
            Tangent vector at `base_point`, initial velocity of the geodesic to
            transport along.
        end_point : array-like, shape=[..., n, n]
            Point to transport to.
            Optional, default: None.
        n_steps : int
            Number of steps to use to approximate the solution of the
            ordinary differential equation.
            Optional, default: 100
        step : str, {'euler', 'rk2', 'rk4'}
            Scheme to use in the integration scheme.
            Optional, default: 'rk4'.

        Returns
        -------
        transported :  array-like, shape=[..., n, n]
            Transported tangent vector at `exp_(base_point)(tangent_vec_b)`.

        References
        ----------
        .. [TP2021] Yann Thanwerdas, Xavier Pennec. O(n)-invariant Riemannian
            metrics on SPD matrices. 2021. ⟨hal-03338601v2⟩

        See Also
        --------
        Integration module: geomstats.integrator
        """
        if end_point is None:
            end_point = self.exp(direction, base_point)

        horizontal_lift_a = gs.linalg.solve_sylvester(
            base_point, base_point, tangent_vec
        )

        square_root_bp, inverse_square_root_bp = powermh(base_point, [0.5, -0.5])
        end_point_lift = Matrices.mul(square_root_bp, end_point, square_root_bp)
        square_root_lift = powermh(end_point_lift, 0.5)

        horizontal_velocity = gs.matmul(inverse_square_root_bp, square_root_lift)
        partial_horizontal_velocity = Matrices.mul(horizontal_velocity, square_root_bp)
        partial_horizontal_velocity = partial_horizontal_velocity + Matrices.transpose(
            partial_horizontal_velocity
        )

        def force(state, time):
            horizontal_geodesic_t = (
                1 - time
            ) * square_root_bp + time * horizontal_velocity
            geodesic_t = (
                (1 - time) ** 2 * base_point
                + time * (1 - time) * partial_horizontal_velocity
                + time**2 * end_point
            )

            align = Matrices.mul(
                horizontal_geodesic_t,
                Matrices.transpose(horizontal_velocity - square_root_bp),
                state,
            )
            right = align + Matrices.transpose(align)
            return gs.linalg.solve_sylvester(geodesic_t, geodesic_t, -right)

        flow = integrate(force, horizontal_lift_a, n_steps=n_steps, step=step)
        final_align = Matrices.mul(end_point, flow[-1])
        return final_align + Matrices.transpose(final_align)

    def injectivity_radius(self, base_point):
        """Compute the upper bound of the injectivity domain.

        This is the smallest eigen value of the base point.

        Parameters
        ----------
        base_point : array-like, shape=[..., n, n]
            Point on the manifold.

        Returns
        -------
        radius : float
            Injectivity radius.
        """
        eigen_values = gs.linalg.eigvalsh(base_point)
        return eigen_values[..., 0] ** 0.5


class SPDEuclideanMetric(MatricesMetric):
    """Class for the Euclidean metric on the SPD manifold."""

    @staticmethod
    def exp_domain(tangent_vec, base_point):
        """Compute the domain of the Euclidean exponential map.

        Compute the real interval of time where the Euclidean geodesic starting
        at point `base_point` in direction `tangent_vec` is defined.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., n, n]
            Tangent vector at base point.
        base_point : array-like, shape=[..., n, n]
            Base point.

        Returns
        -------
        exp_domain : array-like, shape=[..., 2]
            Interval of time where the geodesic is defined.
        """
        invsqrt_base_point = powermh(base_point, -0.5)

        reduced_vec = gs.matmul(invsqrt_base_point, tangent_vec)
        reduced_vec = gs.matmul(reduced_vec, invsqrt_base_point)
        eigvals = gs.linalg.eigvalsh(reduced_vec)
        min_eig = gs.amin(eigvals, axis=-1)
        max_eig = gs.amax(eigvals, axis=-1)

        inf_value = gs.where(max_eig <= 0.0, gs.array(-math.inf), -1.0 / max_eig)
        sup_value = gs.where(min_eig >= 0.0, gs.array(-math.inf), -1.0 / min_eig)

        return gs.stack((inf_value, sup_value), axis=-1)

    def injectivity_radius(self, base_point):
        """Compute the upper bound of the injectivity domain.

        This is the smallest eigen value of the base point.

        Parameters
        ----------
        base_point : array-like, shape=[..., n, n]
            Point on the manifold.

        Returns
        -------
        radius : float
            Injectivity radius.
        """
        eigen_values = gs.linalg.eigvalsh(base_point)
        return eigen_values[..., 0]


class SPDLogEuclideanMetric(PullbackDiffeoMetric):
    """Class for the Log-Euclidean metric on the SPD manifold."""

    def __init__(self, space, image_space=None):
        if image_space is None:
            image_space = SymmetricMatrices(n=space.n)
        diffeo = SymMatrixLog()
        super().__init__(space, diffeo, image_space)


class SPDPowerMetric(PullbackDiffeoMetric):
    r"""Pullback metric induced by the power diffeomorphism.

    Given an equipped image space, the pullback metric is given by

    .. math::

        g_{\Sigma}^{(p)}(X, X)=
        \frac{1}{p^2} g_{\Sigma^p}\left(d_{\Sigma}
        \operatorname{pow}_p(X), d_{\Sigma} \operatorname{pow}_p(X)\right)

    The image space must be equipped with a `ScalarProductMetric`. The scale
    :math:`s` relates with power :math:`p` by

    .. math::

        s = 1 / power^2

    Check section 5.3 of [T2022]_ for more details.

    References
    ----------
    .. [T2022] Thanwerdas, Y. (2022) Riemannian and stratified
        geometries of covariance and correlation matrices. Theses.
        Université Côte d’Azur. Available at:
        https://hal.archives-ouvertes.fr/tel-03698752 (Accessed: 29 September 2022).
    """

    def __init__(self, space, image_space):
        if not isinstance(image_space.metric, ScalarProductMetric):
            raise ValueError(
                "`image-space` must be equipped with a `ScalarProductMetric`"
            )
        power = 1 / gs.sqrt(image_space.metric.scale)
        diffeo = MatrixPower(power)

        super().__init__(space, diffeo, image_space)


class LieCholeskyMetric(PullbackDiffeoMetric):
    """Pullback metric via a diffeomorphism.

    Diffeormorphism between SPD matrices and PLT matrices equipped with
    left invariant metric (see chapter 7 [TP2022]_).

    References
    ----------
    .. [T2022] Yann Thanwerdas. Riemannian and stratified
        geometries on covariance and correlation matrices. Differential
        Geometry [math.DG]. Université Côte d'Azur, 2022.
    """

    def __init__(self, space):
        image_space = PositiveLowerTriangularMatrices(space.n, equip=False)
        image_space.equip_with_metric(InvariantPositiveLowerTriangularMatricesMetric)

        diffeo = CholeskyMap()

        super().__init__(space=space, diffeo=diffeo, image_space=image_space)



import pytest

import geomstats.backend as gs
from geomstats.test.test_case import TestCase
from geomstats.test_cases.geometry.manifold import _ManifoldTestCaseMixins


class ComplexManifoldTestCase(_ManifoldTestCaseMixins, TestCase):
    @pytest.mark.type
    def test_random_point_is_complex(self, n_points):
        point = self.data_generator.random_point(n_points)

        self.assertTrue(gs.is_complex(point))

    @pytest.mark.random
    def test_random_point_imaginary_nonzero(self, n_points, atol):
        point = self.data_generator.random_point(n_points)
        res = gs.imag(gs.abs(point))
        self.assertAllClose(res, gs.zeros_like(point), atol=atol)

    @pytest.mark.type
    def test_random_tangent_vec_is_complex(self, n_points):
        point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(point)

        self.assertTrue(gs.is_complex(tangent_vec))

    @pytest.mark.random
    def test_random_tangent_vec_imaginary_nonzero(self, n_points, atol):
        point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(point)

        res = gs.imag(gs.abs(tangent_vec))
        self.assertAllClose(res, gs.zeros_like(tangent_vec), atol=atol)



import pytest

from geomstats.test_cases.geometry.manifold import ManifoldTestCase


class ProductManifoldTestCase(ManifoldTestCase):
    @pytest.mark.random
    def test_embed_to_product_after_project_from_product(self, n_points, atol):
        point = self.data_generator.random_point(n_points)

        factor_point = self.space.project_from_product(point)
        point_ = self.space.embed_to_product(factor_point)

        self.assertAllClose(point_, point, atol=atol)



import pytest

import geomstats.backend as gs
from geomstats.numerics.finite_differences import forward_difference
from geomstats.test.random import get_random_times
from geomstats.test.vectorization import generate_vectorization_data
from geomstats.test_cases.geometry.fiber_bundle import FiberBundleTestCase
from geomstats.test_cases.geometry.nfold_manifold import NFoldManifoldTestCase
from geomstats.test_cases.geometry.pullback_metric import PullbackDiffeoMetricTestCase
from geomstats.vectorization import get_batch_shape


def elastic_distance(space, point_a, point_b, a=1.0, b=0.5):
    r"""Elastic distance between two points.

    Does not use SRV transform. Implements theorem 2.1 of [BCKKNP2022]_:

    .. math::

        \operatorname{dist}_{a, b}\left(c_1, c_2\right)=
        2 b \sqrt{\ell_{c_1}+\ell_{c_2}-2 \int_D \sqrt{\left|c_1^{\prime}
        \right|\left|c_2^{\prime}\right|} \cos \left(\frac{a}{2 b} \theta\right) d u}

    with

    .. math::

        \theta(u)=\min \left(\cos ^{-1}\left(R\left(c_1\right)
        \cdot R\left(c_2\right) /\left|R\left(c_1\right) \|
        R\left(c_2\right)\right|\right), \frac{2 b \pi}{a}\right)

    Parameters
    ----------
    space : DiscreteCurvesStartingAtOrigin
        Unequipped space.
    point_a : array-like, shape=[..., k_sampling_points - 1, ambient_dim]
        Point.
    point_b : array-like, shape=[..., k_sampling_points - 1, ambient_dim]
        Point.
    a : float
        Bending parameter.
    b : float
        Stretching parameter.

    Returns
    -------
    dist : array-like, shape=[...,]
        Distance.

    References
    ----------
    .. [BCKKNP2022] Martin Bauer, Nicolas Charon, Eric Klassen, Sebastian Kurtek,
        Tom Needham, and Thomas Pierron.
        “Elastic Metrics on Spaces of Euclidean Curves: Theory and Algorithms.”
        arXiv, September 20, 2022. https://doi.org/10.48550/arXiv.2209.09862.
    """
    lambda_ = a / (2 * b)
    k_sampling_points = point_a.shape[-2] + 1

    delta = 1 / (k_sampling_points - 1)

    velocity_a = forward_difference(space.insert_origin(point_a), axis=-2)
    velocity_b = forward_difference(space.insert_origin(point_b), axis=-2)

    length_a = space.length(point_a)
    length_b = space.length(point_b)

    velocity_a_norm = gs.linalg.norm(velocity_a, axis=-1)
    velocity_b_norm = gs.linalg.norm(velocity_b, axis=-1)

    aux_theta = gs.arccos(
        gs.dot(velocity_a, velocity_b) / (velocity_a_norm * velocity_b_norm)
    )
    max_value = gs.pi / lambda_

    theta = gs.where(
        aux_theta > max_value,
        max_value,
        aux_theta,
    )

    integral = gs.sum(
        gs.sqrt(velocity_a_norm * velocity_b_norm) * gs.cos(lambda_ * theta) * delta,
        axis=-1,
    )

    return 2 * b * gs.sqrt(length_a + length_b - 2 * integral)


class DiscreteCurvesStartingAtOriginTestCase(NFoldManifoldTestCase):
    def test_interpolate(self, point, param, expected, atol):
        points = self.space.interpolate(point)(param)
        self.assertAllClose(points, expected, atol=atol)

    def test_interpolate_vec(self, n_reps, n_times, atol):
        point = self.data_generator.random_point()
        param = get_random_times(n_times)

        expected = self.space.interpolate(point)(param)

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    point=point,
                    param=param,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["point"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    def test_length(self, point, expected, atol):
        length = self.space.length(point)
        self.assertAllClose(length, expected, atol=atol)

    def test_normalize(self, point, expected, atol):
        normalized_point = self.space.normalize(point)
        self.assertAllClose(normalized_point, expected, atol)

    @pytest.mark.random
    def test_normalize_is_unit_length(self, n_points, atol):
        point = self.data_generator.random_point(n_points)
        normalized_point = self.space.normalize(point)
        normalize_lengths = self.space.length(normalized_point)
        self.assertAllClose(
            normalize_lengths, gs.ones_like(normalize_lengths), atol=atol
        )


class ReparametrizationBundleTestCase(FiberBundleTestCase):
    @pytest.mark.random
    def test_tangent_vector_projections_orthogonality_with_metric(self, n_points, atol):
        """Test horizontal and vertical projections.

        Check that horizontal and vertical parts of any tangent
        vector are orthogonal with respect to the SRVMetric inner
        product.
        """
        base_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        tangent_vec_hor = self.total_space.fiber_bundle.horizontal_projection(
            tangent_vec, base_point
        )
        tangent_vec_ver = self.total_space.fiber_bundle.vertical_projection(
            tangent_vec, base_point
        )

        res = self.total_space.metric.inner_product(
            tangent_vec_hor, tangent_vec_ver, base_point
        )
        expected_shape = get_batch_shape(self.total_space.point_ndim, base_point)
        expected = gs.zeros(expected_shape)
        self.assertAllClose(res, expected, atol=atol)


class ElasticMetricTestCase(PullbackDiffeoMetricTestCase):
    @pytest.mark.random
    def test_dist_against_no_transform(self, n_points, atol):
        point_a = self.data_generator.random_point(n_points)
        point_b = self.data_generator.random_point(n_points)

        dist = self.space.metric.dist(point_a, point_b)
        dist_ = elastic_distance(
            self.space,
            point_a,
            point_b,
            a=self.space.metric.a,
            b=self.space.metric.b,
        )
        print(point_a.shape, point_b.shape)

        print(dist)
        print(dist_)

        self.assertAllClose(dist, dist_, atol=atol)



import functools
import importlib
import inspect
import json
import os
import shutil
import tempfile
import types
import warnings

import nbformat
import pytest
from matplotlib import pyplot as plt
from nbconvert.preprocessors import ExecutePreprocessor

import geomstats.backend as gs
from geomstats.exceptions import AutodiffNotImplementedError
from geomstats.test.test_case import autodiff_backend
from geomstats.test.vectorization import test_vectorization

NAME2MARK = {
    "xfail": pytest.mark.xfail(),
    "vec": pytest.mark.vec(),
    "skip": pytest.mark.skip(),
}


class TestFunction:
    def __init__(self, name, func, active=False):
        self.name = name
        self.func = func
        self._marks = self._collect_initial_marks()
        self._missing_marks = []
        self.active = active
        self._arg_names = None
        self.tolerances = {}

    def _collect_initial_marks(self):
        if not hasattr(self.func, "pytestmark"):
            return []
        return _get_pytest_mark_names(self.func)

    @property
    def skip(self):
        return "skip" in self.marks

    @property
    def marks(self):
        return self._marks + self._missing_marks

    @property
    def vanilla_name(self):
        return self.name[5:]

    @property
    def data_name(self):
        return f"{self.vanilla_name}_test_data"

    def add_mark(self, mark_name):
        if mark_name not in self.marks:
            self._missing_marks.append(mark_name)

    @property
    def arg_names(self):
        if self._arg_names is None:
            self._arg_names = inspect.getfullargspec(self.func)[0]
            self._handle_tolerances_defaults()
        return self._arg_names

    @property
    def arg_str(self):
        return ", ".join(self.arg_names[1:])

    def _handle_tolerances_defaults(self):
        for arg_name in self.arg_names:
            if arg_name.endswith("rtol"):
                self.tolerances.setdefault(arg_name, gs.rtol)
            elif arg_name.endswith("tol"):
                self.tolerances.setdefault(arg_name, 1e-6)

    def collect(
        self, testing_data, decorators=(), conditional_decorators=(), skip_vec=False
    ):
        test_func, default_values = _copy_func(self.func)
        if not self.active:
            return pytest.mark.ignore()(test_func)

        if self.skip or skip_vec and "vec" in self.marks:
            return pytest.mark.skip()(test_func)

        for decorator in decorators:
            test_func = decorator(test_func)

        for condition_func, decorator in conditional_decorators:
            if condition_func(self):
                test_func = decorator(test_func)

        for mark_name in self._missing_marks:
            test_func = NAME2MARK[mark_name]()(test_func)

        # no args case (note selection was done above)
        if len(self.arg_names) == 1:
            return test_func

        test_data = self._get_test_data(testing_data, default_values)
        return pytest.mark.parametrize(self.arg_str, test_data)(test_func)

    def _get_test_data(self, testing_data, default_values):
        # assumes pairing test-data exists
        test_data = getattr(testing_data, self.data_name)()

        if test_data is None:
            raise Exception(f"'{self.data_name}' returned None. should be list")

        test_data = self._dictify_test_data(test_data)
        test_data = self._pytestify_test_data(test_data, default_values)

        return test_data

    def _dictify_test_data(self, test_data):
        test_data_ = []
        for test_datum in test_data:
            if not isinstance(test_datum, dict):
                test_datum = dict(zip(self.arg_names[1:], test_datum[:-1]))
                test_datum["marks"] = test_datum[-1]

            for tol_arg_name, tol in self.tolerances.items():
                test_datum.setdefault(tol_arg_name, tol)

            test_data_.append(test_datum)

        return test_data_

    def _pytestify_test_data(self, test_data, default_values):
        tests = []
        for test_datum in test_data:
            try:
                values = [
                    test_datum.get(key) if key in test_datum else default_values[key]
                    for key in self.arg_names[1:]
                ]

            except KeyError:
                raise Exception(
                    f"{self.name} requires the following arguments: "
                    f"{', '.join(self.arg_names[1:])}"
                )
            tests.append(pytest.param(*values, marks=test_datum.get("marks")))

        return tests


class AutoVecTestFunction(TestFunction):
    def __init__(self, associated_test_func):
        super().__init__(f"{associated_test_func.name}_vec", None)
        self.associated_test_func = associated_test_func
        self.active = True
        self.add_mark("vec")

    def collect(
        self, testing_data, decorators=(), conditional_decorators=(), skip_vec=False
    ):
        if self.active:
            self._generate_vectorized_func()
        return super().collect(
            testing_data,
            decorators=decorators,
            conditional_decorators=conditional_decorators,
            skip_vec=skip_vec,
        )

    def _generate_vectorized_func(self):
        test_func = self.associated_test_func.func

        def new_test(self, n_reps, atol):
            return test_vectorization(self, test_func, n_reps, atol)

        self.func = new_test


def _update_attrs(test_funcs, testing_data, attrs):
    decorators = _collect_decorators(testing_data)
    conditional_decorators = _collect_conditional_decorators(testing_data)

    for skip_name in testing_data.skips:
        test_funcs[skip_name].add_mark("skip")

    for xfail_name in testing_data.xfails:
        test_funcs[xfail_name].add_mark("xfail")

    for func_name, tolerances in testing_data.tolerances.items():
        test_funcs[func_name].tolerances.update(tolerances)

    for test_func in test_funcs.values():
        attrs[test_func.name] = test_func.collect(
            testing_data,
            decorators=decorators,
            conditional_decorators=conditional_decorators,
            skip_vec=testing_data.skip_vec,
        )

    return attrs


class Parametrizer(type):
    """Metaclass for test classes driven by test definition.

    Note: A test class is a class that inherits from TestCase.
    For example, `class TestEuclidean(TestCase)` defines
    a test class.

    The Parametrizer helps its test class by pairing:
    - the different test functions of the test class:
      - e.g. the test function `test_belongs`,
    - with different test data, generated by auxiliary test data functions
      - e.g. the test data function `belongs_data` that generates data
      to test the function `belongs`.

    As such, Parametrizer acts as a "metaclass" of its test class:
    `class TestEuclidean(TestCase, metaclass=Parametrizer)`.

    Specifically, Parametrizer decorates every test function inside
    its test class with pytest.mark.parametrizer, with the exception
    of the test class' class methods and static methods.

    Two conventions need to be respected:
    1. The test class should contain an attribute named 'testing_data'.
      - `testing_data` is an object inheriting from `TestData`.
    2. Every test function should have its corresponding test data function created
    inside the TestData object called `testing_data`.

    A sample test class looks like this:

    ```
    class TestDataEuclidean(TestData):
        def belongs_data():
            ...
            return self.generate_tests(...)

    class TestEuclidean(TestCase, metaclass=Parametrizer):
        testing_data = TestDataEuclidean()
        def test_belongs():
            ...
    ```
    Parameters
    ----------
    cls : child class of TestCase
        Test class, i.e. a class inheriting from TestCase
    name : str
        Name of the test class
    bases : TestCase
        Parent class of the test class: TestCase.
    attrs : dict
        Attributes of the test class, for example its methods,
        stored in a dictionnary as (key, value) when key gives the
        name of the attribute (for example the name of the method),
        and value gives the actual attribute (for example the method
        itself.)

    References
    ----------
    More on pytest's parametrizers can be found here:
    https://docs.pytest.org/en/6.2.x/parametrize.html
    """

    def __new__(cls, name, bases, attrs):
        testing_data = locals()["attrs"].get("testing_data")
        _raise_missing_testing_data(testing_data)

        test_funcs = _collect_all_tests(attrs, bases, active=True)
        _check_test_data_pairing(test_funcs, testing_data)

        if testing_data.skip_all:
            for test_func in test_funcs.values():
                test_func.add_mark("skip")

        _update_attrs(test_funcs, testing_data, attrs)
        return super().__new__(cls, name, bases, attrs)


class DataBasedParametrizer(type):
    """Metaclass for test classes driven by data definition.

    It differs from `Parametrizer` because every test data function must have
    an associated test function, instead of the opposite.
    """

    def __new__(cls, name, bases, attrs):
        testing_data = locals()["attrs"].get("testing_data")
        _raise_missing_testing_data(testing_data)

        test_funcs = _collect_all_tests(attrs, bases, active=False)

        _activate_tests_given_data(test_funcs, testing_data)

        if testing_data.skip_all:
            for test_func in test_funcs.values():
                test_func.add_mark("skip")

        _update_attrs(test_funcs, testing_data, attrs)
        return super().__new__(cls, name, bases, attrs)


def _get_available_data_names(testing_data):
    return [
        _test_data_name_to_vanilla_name(attr_name)
        for attr_name in dir(testing_data)
        if _is_test_data(attr_name)
    ]


def _raise_missing_testing_data(testing_data):
    if testing_data is None:
        raise Exception("Testing class doesn't have class object named 'testing_data'")


def _collect_decorators(testing_data):
    decorators = []
    if not testing_data.fail_for_autodiff_exceptions and not autodiff_backend():
        decorators.append(_except_autodiff_exception)

    if not testing_data.fail_for_not_implemented_errors:
        decorators.append(_except_not_implemented_errors)

    return decorators


def _trial_condition(test_func):
    marks = test_func.marks
    return "vec" in marks or "random" in marks


def _collect_conditional_decorators(testing_data):
    decorators = []
    if testing_data.trials > 1:
        decorators.append(
            (
                _trial_condition,
                lambda fn: _multiple_trials(fn, trials=testing_data.trials),
            )
        )

    return decorators


def _get_pytest_mark_names(test_func):
    return [elem.name for elem in test_func.pytestmark if isinstance(elem, pytest.Mark)]


def _except_autodiff_exception(func):
    @functools.wraps(func)
    def _wrapped(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except AutodiffNotImplementedError:
            pass

    return _wrapped


def _except_not_implemented_errors(func):
    @functools.wraps(func)
    def _wrapped(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except NotImplementedError:
            pass

    return _wrapped


def _multiple_trials(func, trials):
    @functools.wraps(func)
    def _wrapped(*args, **kwargs):
        trial = 0
        while True:
            try:
                trial += 1
                return func(*args, **kwargs)
            except Exception as exception:
                if trial < trials:
                    continue

                raise exception

    return _wrapped


def _copy_func(
    f,
    name=None,
):
    """Copy function.

    Return a function with same code, globals, defaults, closure, and
    name (or provide a new name).

    Additionally, keyword arguments are transformed into positional arguments for
    compatibility with pytest.
    """
    fn = types.FunctionType(
        f.__code__, f.__globals__, name or f.__name__, f.__defaults__, f.__closure__
    )
    fn.__dict__.update(f.__dict__)

    sign = inspect.signature(fn)
    defaults, new_params = {}, []
    for param in sign.parameters.values():
        if param.default is inspect._empty:
            new_params.append(param)
        else:
            new_params.append(inspect.Parameter(param.name, kind=1))
            defaults[param.name] = param.default
    new_sign = sign.replace(parameters=new_params)
    fn.__signature__ = new_sign

    return fn, defaults


def _is_test(attr_name):
    return attr_name.startswith("test_")


def _is_test_data(attr_name):
    return attr_name.endswith("_test_data")


def _test_data_name_to_vanilla_name(test_data_name):
    return test_data_name[:-10]


def _collect_available_tests(attrs):
    return {attr_name: attr for attr_name, attr in attrs.items() if _is_test(attr_name)}


def _collect_available_base_tests(bases):
    base_tests = dict()
    for base in bases:
        base_test_names = [name for name in dir(base) if _is_test(name)]
        for test_name in base_test_names:
            if test_name not in base_tests:
                base_tests[test_name] = getattr(base, test_name)

    return base_tests


def _collect_all_tests(attrs, bases, active=True):
    test_attrs = _collect_available_tests(attrs)
    base_attrs = _collect_available_base_tests(bases)

    # order matters
    tests = {**base_attrs, **test_attrs}

    all_tests = {}
    for name, func in tests.items():
        test_function = TestFunction(name, func, active=active)
        all_tests[test_function.vanilla_name] = test_function

    return all_tests


def _raise_missing_tests(missing_tests):
    if missing_tests:
        msg = "Need to define tests for:"
        for name in set(missing_tests):
            msg += f"\n\t-{name}_test_data"

        raise Exception(msg)


def _activate_tests_given_data(test_funcs, testing_data):
    defined_names_ls = _get_available_data_names(testing_data)

    missing_tests = []
    for name in defined_names_ls:
        if name in test_funcs:
            test_funcs[name].active = True

        elif not name.endswith("_vec"):
            missing_tests.append(name)
            continue

        else:
            name_no_vec = name[:-4]
            if name_no_vec not in test_funcs:
                missing_tests.append(name)
                continue

            test_function = AutoVecTestFunction(test_funcs[name_no_vec])
            test_funcs[test_function.vanilla_name] = test_function

    _raise_missing_tests(missing_tests)

    return test_funcs


def _raise_missing_data(missing_data):
    if missing_data:
        msg = "Need to define data for:"
        for name in set(missing_data):
            msg += f"\n\t-test_{name}"

        raise Exception(msg)


def _check_test_data_pairing(test_funcs, testing_data):
    defined_names_ls = _get_available_data_names(testing_data)
    missing_data = []
    for name in test_funcs.keys():
        if name not in defined_names_ls:
            missing_data.append(name)

    _raise_missing_data(missing_data)

    return test_funcs


def _run_example(path, **kwargs):
    warnings.simplefilter("ignore", category=UserWarning)

    spec = importlib.util.spec_from_file_location("module.name", path)
    example = importlib.util.module_from_spec(spec)

    spec.loader.exec_module(example)

    example.main(**kwargs)
    plt.close("all")


class ExamplesParametrizer(type):
    def __new__(cls, name, bases, attrs):
        def _create_new_test(path, **kwargs):
            def new_test(self):
                return _run_example(path=path, **kwargs)

            return new_test

        BACKEND = os.environ.get("GEOMSTATS_BACKEND", "numpy")

        testing_data = locals()["attrs"].get("testing_data")
        _raise_missing_testing_data(testing_data)

        paths = testing_data.paths
        metadata = testing_data.metadata
        skips = testing_data.skips

        for path in paths:
            name = path.split(os.sep)[-1].split(".")[0]

            func_name = f"test_{name}"

            metadata_ = metadata.get(name, {})
            if not isinstance(metadata_, dict):
                metadata_ = {"backends": metadata_}

            kwargs = metadata_.get("kwargs", {})
            test_func = _create_new_test(path, **kwargs)

            backends = metadata_.get("backends", None)
            if name in skips or (backends and BACKEND not in backends):
                test_func = pytest.mark.skip()(test_func)

            attrs[func_name] = test_func

        return super().__new__(cls, name, bases, attrs)


def _exec_notebook(path):
    file_name = tempfile.NamedTemporaryFile(suffix=".ipynb").name
    shutil.copy(path, file_name)

    with open(file_name) as file:
        notebook = nbformat.read(file, as_version=4)

    eprocessor = ExecutePreprocessor(timeout=1000, kernel_name="python3")
    eprocessor.preprocess(notebook)


class NotebooksParametrizer(type):
    def __new__(cls, name, bases, attrs):
        def _create_new_test(path, **kwargs):
            def new_test(self):
                return _exec_notebook(path=path)

            return new_test

        BACKEND = os.environ.get("GEOMSTATS_BACKEND", "numpy")

        testing_data = locals()["attrs"].get("testing_data")
        _raise_missing_testing_data(testing_data)

        paths = testing_data.paths

        for path in paths:
            name = path.split(os.sep)[-1].split(".")[0]

            func_name = f"test_{name}"
            test_func = _create_new_test(path)

            with open(path, "r", encoding="utf8") as file:
                metadata = json.load(file).get("metadata")

            backends = metadata.get("backends", None)
            if backends and BACKEND not in backends:
                test_func = pytest.mark.skip()(test_func)

            attrs[func_name] = test_func

        return super().__new__(cls, name, bases, attrs)



"""Visualization for Geometric Statistics."""

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D  # NOQA

import geomstats.backend as gs
from geomstats.geometry.hyperboloid import Hyperboloid
from geomstats.geometry.poincare_half_space import PoincareHalfSpace

H2 = Hyperboloid(dim=2)
POINCARE_HALF_PLANE = PoincareHalfSpace(dim=2)


AX_SCALE = 1.2


class KleinDisk:
    def __init__(self, points=None):
        self.center = gs.array([0.0, 0.0])
        self.points = []
        if points is not None:
            self.add_points(points)

    @staticmethod
    def set_ax(ax=None):
        """Set axis."""
        if ax is None:
            ax = plt.subplot()
        ax_s = AX_SCALE
        plt.setp(ax, xlim=(-ax_s, ax_s), ylim=(-ax_s, ax_s), xlabel="X", ylabel="Y")
        return ax

    def add_points(self, points):
        if not gs.all(H2.belongs(points)):
            raise ValueError("Points do not belong to the hyperbolic space.")
        points = self.convert_to_klein_coordinates(points)
        if not isinstance(points, list):
            points = list(points)
        self.points.extend(points)

    @staticmethod
    def convert_to_klein_coordinates(points):
        poincare_coords = points[:, 1:] / (1 + points[:, :1])
        poincare_radius = gs.linalg.norm(poincare_coords, axis=1)
        poincare_angle = gs.arctan2(poincare_coords[:, 1], poincare_coords[:, 0])

        klein_radius = 2 * poincare_radius / (1 + poincare_radius**2)
        klein_angle = poincare_angle

        coords_0 = gs.expand_dims(klein_radius * gs.cos(klein_angle), axis=1)
        coords_1 = gs.expand_dims(klein_radius * gs.sin(klein_angle), axis=1)
        klein_coords = gs.concatenate([coords_0, coords_1], axis=1)
        return klein_coords

    def draw(self, ax, **kwargs):
        circle = plt.Circle((0, 0), radius=1.0, color="black", fill=False)
        ax.add_artist(circle)
        points_x = [gs.to_numpy(point[0]) for point in self.points]
        points_y = [gs.to_numpy(point[1]) for point in self.points]
        ax.scatter(points_x, points_y, **kwargs)

    def plot(self, points, ax=None, **point_draw_kwargs):
        ax = self.set_ax(ax=ax)
        self.points = []
        self.add_points(points)
        self.draw(ax, **point_draw_kwargs)


class PoincareDisk:
    def __init__(self, points=None, coords_type="extrinsic"):
        self.center = gs.array([0.0, 0.0])
        self.points = []
        self.coords_type = coords_type
        if points is not None:
            self.add_points(points)

    @staticmethod
    def set_ax(ax=None):
        """Set axis."""
        if ax is None:
            ax = plt.subplot()
        ax_s = AX_SCALE
        plt.setp(
            ax,
            xlim=(-ax_s, ax_s),
            ylim=(-ax_s, ax_s),
            xlabel="X",
            ylabel="Y",
        )
        return ax

    def add_points(self, points):
        if self.coords_type == "extrinsic":
            if not gs.all(H2.belongs(points)):
                raise ValueError("Points do not belong to the hyperbolic space.")
            points = self.convert_to_poincare_coordinates(points)

        if not isinstance(points, list):
            points = list(points)

        if gs.all([len(point) == 2 for point in self.points]):
            self.points.extend(points)
        else:
            raise ValueError("Points do not have dimension 2.")

    @staticmethod
    def convert_to_poincare_coordinates(points):
        poincare_coords = points[:, 1:] / (1 + points[:, :1])
        return poincare_coords

    def draw(self, ax, **kwargs):
        circle = plt.Circle((0, 0), radius=1.0, color="black", fill=False)
        ax.add_artist(circle)
        if len(self.points) > 0:
            if gs.all([len(point) == 2 for point in self.points]):
                points_x = gs.stack([point[0] for point in self.points], axis=0)
                points_y = gs.stack([point[1] for point in self.points], axis=0)
                ax.scatter(points_x, points_y, **kwargs)
            else:
                raise ValueError("Points do not have dimension 2.")

    def plot(self, points, ax=None, coords_type=None, **point_draw_kwargs):
        if coords_type is None:
            coords_type = "extrinsic"
        poincare_disk = PoincareDisk(coords_type=coords_type)
        ax = poincare_disk.set_ax(ax=ax)
        self.points = []
        poincare_disk.add_points(points)
        poincare_disk.draw(ax, **point_draw_kwargs)
        plt.axis("off")


class PoincareHalfPlane:
    """Class used to plot points in the Poincare Half Plane."""

    def __init__(self, points=None, coords_type="half-space"):
        self.points = []
        self.coords_type = coords_type
        if points is not None:
            self.add_points(points)

    def add_points(self, points):
        if self.coords_type == "extrinsic":
            if not gs.all(H2.belongs(points)):
                raise ValueError(
                    "Points do not belong to the hyperbolic space "
                    "(extrinsic coordinates)"
                )
            points = self.convert_to_half_plane_coordinates(points)
        elif self.coords_type == "half-space":
            if not gs.all(POINCARE_HALF_PLANE.belongs(points)):
                raise ValueError(
                    "Points do not belong to the hyperbolic space "
                    "(Poincare half plane coordinates)."
                )
        if not isinstance(points, list):
            points = list(points)
        self.points.extend(points)

    def set_ax(self, ax=None):
        """Set axis."""
        if ax is None:
            ax = plt.subplot()

        plt.setp(ax, xlabel="X", ylabel="Y")
        return ax

    @staticmethod
    def convert_to_half_plane_coordinates(points):
        disk_coords = points[:, 1:] / (1 + points[:, :1])
        disk_x = disk_coords[:, 0]
        disk_y = disk_coords[:, 1]

        denominator = disk_x**2 + (1 - disk_y) ** 2
        coords_0 = gs.expand_dims(2 * disk_x / denominator, axis=1)
        coords_1 = gs.expand_dims((1 - disk_x**2 - disk_y**2) / denominator, axis=1)

        half_plane_coords = gs.concatenate([coords_0, coords_1], axis=1)
        return half_plane_coords

    def draw(self, ax, **kwargs):
        points_x = [gs.to_numpy(point[0]) for point in self.points]
        points_y = [gs.to_numpy(point[1]) for point in self.points]
        ax.scatter(points_x, points_y, **kwargs)

    def plot(self, points, ax=None, coords_type=None, **point_draw_kwargs):
        if coords_type is None:
            coords_type = "half-space"
        poincare_half_plane = PoincareHalfPlane(coords_type=coords_type)
        ax = poincare_half_plane.set_ax(ax=ax)
        self.points = []
        poincare_half_plane.add_points(points)
        poincare_half_plane.draw(ax, **point_draw_kwargs)



"""The Poincare polydisk.

The Poincare polydisk is defined as a product manifold of the Hyperbolic space
of dimension 2. The Poincare polydisk has a product metric. The metric on each
space is the natural Poincare metric multiplied by a constant.

Lead author: Yann Cabanes.

References
----------
.. [JV2016] B. Jeuris and R. Vandebril. The Kahler mean of Block-Toeplitz
    matrices with Toeplitz structured blocks, 2016.
    https://epubs.siam.org/doi/pdf/10.1137/15M102112X
"""

import geomstats.backend as gs
from geomstats.geometry._hyperbolic import _Hyperbolic
from geomstats.geometry.hyperboloid import Hyperboloid
from geomstats.geometry.nfold_manifold import NFoldManifold, NFoldMetric


class PoincarePolydisk(NFoldManifold):
    r"""Class for the Poincare polydisk.

    The Poincare polydisk is a direct product of n Poincare disks,
    i.e. hyperbolic spaces of dimension 2.

    Parameters
    ----------
    n_disks : int
        Number of disks.
    """

    def __init__(self, n_disks, equip=True):
        self.n_disks = n_disks
        super().__init__(base_manifold=Hyperboloid(2), n_copies=n_disks, equip=equip)

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return PoincarePolydiskMetric

    @staticmethod
    def intrinsic_to_extrinsic_coords(point_intrinsic):
        """Convert point from intrinsic to extrensic coordinates.

        Convert the parameterization of a point in the hyperbolic space
        from its intrinsic coordinates, to its extrinsic coordinates
        in Minkowski space.

        Parameters
        ----------
        point_intrinsic : array-like, shape=[..., n_disk, dim]
            Point in intrinsic coordinates.

        Returns
        -------
        point_extrinsic : array-like, shape=[..., n_disks, dim + 1]
            Point in extrinsic coordinates.
        """
        n_disks = point_intrinsic.shape[1]
        point_extrinsic = gs.stack(
            [
                _Hyperbolic.change_coordinates_system(
                    point_intrinsic[:, i_disk, ...], "intrinsic", "extrinsic"
                )
                for i_disk in range(n_disks)
            ],
            axis=1,
        )
        return point_extrinsic


class PoincarePolydiskMetric(NFoldMetric):
    r"""Class defining the Poincare polydisk metric.

    The Poincare polydisk metric is a product of n Poincare metrics,
    each of them being multiplied by a specific constant factor (see
    [JV2016]_).

    This metric comes from a model used to represent
    stationary complex autoregressive Gaussian signals.


    References
    ----------
    .. [JV2016] B. Jeuris and R. Vandebril. The Kähler mean of Block-Toeplitz
        matrices with Toeplitz structured blocks, 2016.
        https://epubs.siam.org/doi/pdf/10.1137/15M102112X
    """

    def __init__(self, space):
        scales = gs.array(
            [float(space.n_disks - i_disk) for i_disk in range(space.n_disks)]
        )
        super().__init__(space, scales=scales)
